/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.eingang.enterprise.entry;

import java.io.IOException;
import java.io.InputStream;
import java.time.ZonedDateTime;
import java.util.Collection;
import java.util.Objects;

import org.apache.commons.collections4.CollectionUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.core.io.Resource;
import org.springframework.http.HttpStatus;
import org.springframework.http.MediaType;
import org.springframework.stereotype.Controller;
import org.springframework.web.bind.annotation.PostMapping;
import org.springframework.web.bind.annotation.RequestMapping;
import org.springframework.web.bind.annotation.RequestPart;
import org.springframework.web.bind.annotation.ResponseBody;
import org.springframework.web.bind.annotation.ResponseStatus;
import org.springframework.web.multipart.MultipartFile;

import de.ozgcloud.common.binaryfile.TempFileUtils;
import de.ozgcloud.common.errorhandling.TechnicalException;
import de.ozgcloud.eingang.common.formdata.FormData;
import de.ozgcloud.eingang.common.formdata.IncomingFile;
import de.ozgcloud.eingang.common.formdata.IncomingFileGroup;
import de.ozgcloud.eingang.common.vorgang.VorgangNummerSupplier;
import de.ozgcloud.eingang.enterprise.entry.EntryResponse.ResponseVorgang;
import de.ozgcloud.eingang.semantik.SemantikAdapter;

@Controller
@ResponseBody
@RequestMapping("antrag")
public class EntryController {

	private static final String STARTING_STATUS = "NEU";
	private static final String OTHER_FILE_GROUP_NAME = "Sonstige";

	@Autowired
	private EntryDataMapper mapper;

	@Autowired
	private SemantikAdapter semantikAdapter;
	@Autowired
	private VorgangNummerSupplier vorgangNummerSupplier;

	@ResponseStatus(HttpStatus.ACCEPTED)
	@PostMapping(consumes = "multipart/form-data", produces = MediaType.APPLICATION_JSON_VALUE)
	public EntryResponse receiveAntrag(@RequestPart("formData") Resource formData,
			@RequestPart(name = "representation", required = false) Collection<MultipartFile> representations,
			@RequestPart(name = "attachment", required = false) Collection<MultipartFile> attachment)
			throws IOException {
		var mapped = mapper.mapEntryData(formData.getInputStream());
		mapped = addVorgangNummer(mapped);
		mapped = addRepresentations(representations, mapped);
		mapped = addAttachments(attachment, mapped);

		var vorgangId = semantikAdapter.processFormData(mapped);

		return buildResponse(mapped, vorgangId);
	}

	private FormData addVorgangNummer(FormData formData) {
		var header = formData.getHeader().toBuilder().vorgangNummer(vorgangNummerSupplier.get()).build();
		return formData.toBuilder().header(header).build();
	}

	private FormData addRepresentations(Collection<MultipartFile> representations, FormData formData) {
		if (Objects.isNull(representations)) {
			return formData;
		}

		var builder = formData.toBuilder();
		representations.stream().map(this::buildIncomingFile).forEach(builder::representation);
		return builder.numberOfRepresentations(representations.size()).build();
	}

	FormData addAttachments(Collection<MultipartFile> attachments, FormData formData) {
		if (CollectionUtils.isEmpty(attachments)) {
			return formData;
		}

		IncomingFileGroup group = buildSonstigeGroup(attachments);
		return formData.toBuilder().attachment(group).numberOfAttachments(group.getFiles().size()).build();
	}

	private IncomingFileGroup buildSonstigeGroup(Collection<MultipartFile> attachments) {
		var builder = IncomingFileGroup.builder().name(OTHER_FILE_GROUP_NAME);
		attachments.stream().map(this::buildIncomingFile).forEach(builder::file);
		return builder.build();
	}

	private IncomingFile buildIncomingFile(MultipartFile multipartFile) {
		InputStream inStream = getInputStream(multipartFile);

		return IncomingFile.builder()
				.name(multipartFile.getOriginalFilename())
				.contentType(multipartFile.getContentType())
				.size(multipartFile.getSize())
				.file(TempFileUtils.writeTmpFile(inStream))
				.build();
	}

	private InputStream getInputStream(MultipartFile multipartFile) {
		try {
			return multipartFile.getInputStream();
		} catch (IOException e) {
			throw new TechnicalException("Error reading incoming multipart File.", e);
		}
	}

	EntryResponse buildResponse(FormData formData, String vorgangId) {
		return EntryResponse.builder()
				.transactionId(formData.getHeader().getRequestId())
				.vorgang(ResponseVorgang.builder()
						.vorgangId(vorgangId)
						.vorgangNummer(formData.getHeader().getVorgangNummer())
						.status(STARTING_STATUS)
						.statusSince(ZonedDateTime.now().withNano(0))
						.build())
				.build();
	}
}
