/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.eingang.enterprise.entry;

import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.stream.Collectors;

import org.apache.commons.lang3.tuple.Pair;
import org.mapstruct.Mapper;
import org.mapstruct.Mapping;

import de.ozgcloud.eingang.common.formdata.FormData;
import de.ozgcloud.eingang.common.formdata.FormData.Representations;
import de.ozgcloud.eingang.common.formdata.FormHeader;
import de.ozgcloud.eingang.common.formdata.PostfachAddressIdentifier;
import de.ozgcloud.eingang.common.formdata.ServiceKonto.PostfachAddress;
import de.ozgcloud.eingang.common.formdata.StringBasedIdentifier;
import de.ozgcloud.eingang.common.formdata.ZustaendigeStelle;

@Mapper
public interface FormDataMapper {

	public static final String VALUE_KEY = "value";
	public static final String LABEL_KEY = "label";

	@Mapping(target = "antragsteller", ignore = true)
	@Mapping(target = "attachment", ignore = true)
	@Mapping(target = "attachments", ignore = true)
	@Mapping(target = "numberOfAttachments", ignore = true)
	@Mapping(target = "representation", ignore = true)
	@Mapping(target = "representations", ignore = true)
	@Mapping(target = "numberOfRepresentations", ignore = true)

	@Mapping(target = "id", ignore = true)
	@Mapping(target = "header", source = "control")

	@Mapping(target = "zustaendigeStelle", source = "control.zustaendigeStelle")
	FormData mapEntryData(EntryData entryData);

	default <T> Optional<T> mapOptional(T obj) {
		return Optional.ofNullable(obj);
	}

	@Mapping(target = "vorgangNummer", ignore = true)
	@Mapping(target = "createdAt", ignore = true)
	@Mapping(target = "channel", constant = "B2G")
	@Mapping(target = "formEngineName", constant = "A12") // TODO
	@Mapping(target = "requestId", source = "transactionId")
	@Mapping(target = "sender", ignore = true) // TODO fill from authentication
	@Mapping(target = "serviceKonto.postfachAddresses", ignore = true)
	FormHeader mapHeader(EntryData.ControlData controlData);

	@Mapping(target = "primaryFormDataPdfRepresentation", ignore = true)
	@Mapping(target = "primaryFormDataRepresentation", source = "formData")
	Representations mapRepresentations(EntryData.ControlData.Representations representations);

	default ZustaendigeStelle fromId(String organisationsEinheitenId) {
		return ZustaendigeStelle.builder().organisationseinheitenId(organisationsEinheitenId).build();
	}

	default Map<String, Object> mapFormItems(List<EntryFormDataItem> items) {
		return items.stream().map(item -> Pair.of(item.getName(),
				item.isFormField() ? mapFormField((EntryFormDataField) item) : mapSubForm((EntryFormDataSubForm) item)))
				.collect(Collectors.toMap(Pair::getKey, Pair::getValue));
	}

	default Map<String, Object> mapFormField(EntryFormDataField field) {
		var map = new HashMap<String, Object>();
		map.put(LABEL_KEY, field.getLabel());
		map.put(VALUE_KEY, field.getValue());

		return Collections.unmodifiableMap(map);
	}

	default Map<String, Object> mapSubForm(EntryFormDataSubForm subForm) {
		var map = new HashMap<String, Object>();
		map.put(LABEL_KEY, subForm.getLabel());
		map.put(VALUE_KEY, mapFormItems(subForm.getFormItems()));

		return Collections.unmodifiableMap(map);
	}

	default PostfachAddress map(de.ozgcloud.eingang.enterprise.entry.EntryData.PostfachAddress address) {
		return PostfachAddress.builder()
				.version("1")
				.type(toNumericType(address.getType()))
				.identifier(StringBasedIdentifier.builder().postfachId(address.getIdentifier()).build())
				.build();
	}

	default int toNumericType(String type) {
		switch (type) {
		case "privat":
			return 0;
		case "unternehmen":
			return 1;
		case "behoerde":
			return 2;
		default:
			return -1;
		}
	}

	default PostfachAddressIdentifier map(String value) {
		return StringBasedIdentifier.builder().postfachId(value).build();
	}

}
