/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.eingang.enterprise.entry;

import static org.assertj.core.api.Assertions.*;
import static org.assertj.core.api.InstanceOfAssertFactories.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;
import static org.springframework.test.web.servlet.request.MockMvcRequestBuilders.*;
import static org.springframework.test.web.servlet.result.MockMvcResultMatchers.*;

import java.io.InputStream;
import java.util.Collections;
import java.util.List;

import org.apache.commons.io.IOUtils;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.ArgumentCaptor;
import org.mockito.Captor;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;
import org.springframework.core.io.Resource;
import org.springframework.mock.web.MockMultipartFile;
import org.springframework.test.web.servlet.MockMvc;
import org.springframework.test.web.servlet.ResultActions;
import org.springframework.test.web.servlet.setup.MockMvcBuilders;

import de.ozgcloud.common.test.TestUtils;
import de.ozgcloud.eingang.common.formdata.FormData;
import de.ozgcloud.eingang.common.formdata.FormDataTestFactory;
import de.ozgcloud.eingang.common.formdata.FormHeader;
import de.ozgcloud.eingang.common.formdata.FormHeaderTestFactory;
import de.ozgcloud.eingang.common.formdata.IncomingFileGroup;
import de.ozgcloud.eingang.common.vorgang.VorgangNummerSupplier;
import de.ozgcloud.eingang.semantik.SemantikAdapter;
import lombok.SneakyThrows;

class EntryControllerTest {

	@Spy
	@InjectMocks
	private EntryController controller;

	@Mock
	private EntryDataMapper mapper;
	@Mock
	private SemantikAdapter semantikAdapter;
	@Mock
	private VorgangNummerSupplier vorgangNummerSupplier;

	private MockMvc mockMvc;

	@BeforeEach
	void init() {
		mockMvc = MockMvcBuilders.standaloneSetup(controller).build();
	}

	@Nested
	class ReceiveAntrag {

		private final FormHeader header = FormHeaderTestFactory.createBuilder().vorgangNummer(null).build();
		private final FormData formData = FormDataTestFactory.createBuilder().header(header).clearAttachments().clearRepresentations().build();

		@Captor
		private ArgumentCaptor<InputStream> streamCaptor;
		@Captor
		private ArgumentCaptor<FormData> formDataCaptor;

		private final EntryResponse response = EntryResponseTestFactory.create();

		@BeforeEach
		void init() {
			when(mapper.mapEntryData(any())).thenReturn(formData);
			doReturn(response).when(controller).buildResponse(any(), any());
			when(vorgangNummerSupplier.get()).thenReturn(FormHeaderTestFactory.VORGANG_NUMMER);
		}

		@Test
		@SneakyThrows
		void shouldReturnAccepted() {
			doPostRequest().andExpect(status().isAccepted());
		}

		@Test
		@SneakyThrows
		void shouldCallMapper() {
			var request = TestUtils.loadTextFile("request/simple.json");

			doPostRequest();

			verify(mapper).mapEntryData(streamCaptor.capture());
			var inputBytes = IOUtils.readFully(streamCaptor.getValue(), request.getBytes().length);
			assertThat(inputBytes).hasSameSizeAs(request.getBytes()).isEqualTo(request.getBytes());
		}

		@Test
		void shouldCallSemantikAdapter() {
			doPostRequest();

			verify(semantikAdapter).processFormData(notNull());
		}

		@Test
		void shouldSetVorgangNummer() {
			doPostRequest();

			verify(semantikAdapter).processFormData(formDataCaptor.capture());
			assertThat(formDataCaptor.getValue().getHeader().getVorgangNummer()).isEqualTo(FormHeaderTestFactory.VORGANG_NUMMER);
		}

		@Test
		@SneakyThrows
		void shouldReturnResponse() {
			var response = controller.receiveAntrag(mock(Resource.class), Collections.emptyList(), Collections.emptyList());

			assertThat(response).isSameAs(this.response);
		}

		@SneakyThrows
		private ResultActions doPostRequest() {
			return mockMvc.perform(multipart("/antrag")
					.file(new MockMultipartFile("formData", TestUtils.loadTextFile("request/simple.json").getBytes())))
					.andExpect(status().is2xxSuccessful());
		}

		@Nested
		class WithRepresentations {

			@Captor
			private ArgumentCaptor<FormData> formDataCaptor;

			@Test
			void shouldAddRepresentations() {
				var formData = doPostRequest();

				assertThat(formData.getRepresentations()).hasSize(2);
			}

			@Test
			void shouldSetNumberOfRepresentations() {
				var formData = doPostRequest();

				assertThat(formData.getNumberOfRepresentations()).isEqualTo(2);
			}

			@SneakyThrows
			private FormData doPostRequest() {
				mockMvc.perform(multipart("/antrag")
						.file(new MockMultipartFile("formData", TestUtils.loadTextFile("request/a12_entry.json").getBytes()))
						.file(new MockMultipartFile("representation", TestUtils.loadTextFile("request/document_JaS.json").getBytes()))
						.file(new MockMultipartFile("representation",
								TestUtils.loadTextFile("request/jugendsozialarbeit_an_schulen-DM.json").getBytes())))
						.andExpect(status().is2xxSuccessful());

				verify(semantikAdapter).processFormData(formDataCaptor.capture());
				return formDataCaptor.getValue();
			}
		}

		@Nested
		class WithAttachments {
			@Test
			void shouldAddAttachments() {
				var formData = doPostRequest();

				assertThat(formData.getAttachments()).hasSize(1).first().extracting(IncomingFileGroup::getFiles).asInstanceOf(LIST).hasSize(1);
			}

			@Test
			void shouldSetNumberOfAttachments() {
				var formData = doPostRequest();

				assertThat(formData.getNumberOfAttachments()).isEqualTo(1);
			}

			@SneakyThrows
			private FormData doPostRequest() {
				mockMvc.perform(multipart("/antrag")
						.file(new MockMultipartFile("formData", TestUtils.loadTextFile("request/a12_entry.json").getBytes()))
						.file(new MockMultipartFile("attachment", TestUtils.loadTextFile("request/document_JaS.json").getBytes())))
						.andExpect(status().is2xxSuccessful());

				verify(semantikAdapter).processFormData(formDataCaptor.capture());
				return formDataCaptor.getValue();
			}

		}

	}

	@Nested
	class TestAddAttachments {

		private MockMultipartFile file = new MockMultipartFile("attachment", TestUtils.loadTextFile("request/document_JaS.json").getBytes());

		@Test
		void shouldIgnoreNullCollection() {
			var formData = FormDataTestFactory.create();

			var result = controller.addAttachments(null, formData);

			assertThat(result).isSameAs(formData);
		}

		@Test
		void shouldIgnoreEmptyCollection() {
			var formData = FormDataTestFactory.create();

			var result = controller.addAttachments(Collections.emptyList(), formData);

			assertThat(result).isSameAs(formData);
		}

		@Test
		void shouldAddToSonstigeGroup() {
			var formData = FormDataTestFactory.createBuilder().clearAttachments().build();

			var result = controller.addAttachments(List.of(file), formData);

			assertThat(result.getAttachments()).hasSize(1).first().extracting(IncomingFileGroup::getName).isEqualTo("Sonstige");
		}

		@Test
		void shouldAddFiles() {
			var formData = FormDataTestFactory.createBuilder().clearAttachments().build();

			var result = controller.addAttachments(List.of(file), formData);

			assertThat(result.getAttachments()).hasSize(1).first().extracting(IncomingFileGroup::getFiles).asInstanceOf(LIST).hasSize(1);
		}
	}

	@Nested
	class BuildResponse {

		@Test
		@SneakyThrows
		void shouldCreateResponse() {
			var response = controller.buildResponse(FormDataTestFactory.create(), ResponseVorgangTestFactory.VORGANG_ID);

			assertThat(response).usingRecursiveComparison()
					.ignoringFields("vorgang.statusSince")
					.isEqualTo(EntryResponseTestFactory.create());
		}
	}

}
