/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.eingang.fim;

import java.util.Collections;
import java.util.List;
import java.util.Map;

import org.springframework.stereotype.Component;
import org.w3c.dom.Document;

import de.ozgcloud.eingang.common.formdata.FormData;
import de.ozgcloud.eingang.common.formdata.IncomingFile;
import de.ozgcloud.eingang.fim.common.xml.DocumentHelper;
import de.ozgcloud.eingang.semantik.enginebased.EngineBasedSemantikAdapter;
import lombok.RequiredArgsConstructor;
import lombok.extern.log4j.Log4j2;

@Log4j2
@RequiredArgsConstructor
@Component
public class FimBasedAdapter implements EngineBasedSemantikAdapter {

	static final String FIM_FORM_ENGINE_NAME = "FIM";

	private final List<FimEngineBasedMapper> mappers;
	private final FimService service;

	@Override
	public boolean isResponsible(FormData formData) {
		return FIM_FORM_ENGINE_NAME.equals(formData.getHeader().getFormEngineName());
	}

	@Override
	public FormData parseFormData(FormData formData) {
		var processedFormData = addFormDataDataMap(formData);

		for (var mapper : mappers) {
			processedFormData = mapper.parseFormData(processedFormData);
		}
		return processedFormData;
	}

	FormData addFormDataDataMap(FormData formData) {
		return formData.toBuilder().formData(parseRequestData(formData)).build();
	}

	Map<String, Object> parseRequestData(FormData formData) {
		return service.findPrimaryRepresentation(formData).map(this::doParse).orElseGet(Collections::emptyMap);
	}

	Map<String, Object> doParse(IncomingFile file) {
		try {
			return processData(DocumentHelper.parse(file.getContentStream()));
		} catch (Exception e) {
			LOG.warn("Error on parsing Document.", e);
			return Collections.emptyMap();
		}
	}

	Map<String, Object> processData(Document content) {
		return FimDocumentHelper.createProcessor(content, service.getScheme(content)).process();
	}
}