/*
 * Copyright (C) 2025 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.eingang.fim;

import java.util.Collections;
import java.util.LinkedHashMap;
import java.util.Map;
import java.util.Optional;

import org.springframework.stereotype.Component;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;

import de.ozgcloud.eingang.fim.common.xml.DocumentHelper;
import lombok.AccessLevel;
import lombok.NoArgsConstructor;
import lombok.RequiredArgsConstructor;
import lombok.extern.log4j.Log4j2;

@NoArgsConstructor(access = AccessLevel.PRIVATE)
@Component
class FimDocumentHelper {

	public static FimDocumentProcessor createProcessor(Document doc, FimScheme fimScheme) {
		return new FimDocumentProcessor(doc, fimScheme);
	}

	@Log4j2
	@RequiredArgsConstructor(access = AccessLevel.PRIVATE)
	static class FimDocumentProcessor {

		private static final String LABEL_KEY = "label";
		private static final String VALUE_KEY = "value";

		private final Document document;
		private final FimScheme scheme;

		public Map<String, Object> process() {
			var data = new LinkedHashMap<String, Object>();
			processNode(document.getDocumentElement(), data);
			return Collections.unmodifiableMap(data);
		}

		void processNode(Node element, Map<String, Object> data) {
			LOG.debug(() -> "process node... " + element.getNodeName());
			DocumentHelper.getChildNodes(element.getChildNodes()).forEach(childElement -> processElement(childElement, data));
		}

		void processElement(Element element, Map<String, Object> data) {
			if (DocumentHelper.isTextNode(element)) {
				addToMap(element, data, element.getTextContent());
			} else {
				var childMap = new LinkedHashMap<String, Object>();
				addToMap(element, data, childMap);
				processNode(element, childMap);
			}
		}

		private void addToMap(Element element, Map<String, Object> data, Object value) {
			var fieldName = DocumentHelper.getFieldNameWithoutNamespace(element.getNodeName());
			data.put(fieldName, buildFieldMap(getFieldName(element).orElse(fieldName), value));
		}

		private Optional<String> getFieldName(Element element) {
			return scheme.getFieldName(element.getNodeName());
		}

		private Map<String, Object> buildFieldMap(String fieldName, Object obj) {
			return Map.of(LABEL_KEY, fieldName, VALUE_KEY, obj);
		}
	}
}