/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.eingang.fim;

import java.io.File;
import java.io.IOException;
import java.util.LinkedHashMap;

import javax.xml.parsers.ParserConfigurationException;

import org.springframework.core.io.ResourceLoader;
import org.w3c.dom.Document;
import org.xml.sax.SAXException;

import de.ozgcloud.eingang.common.errorhandling.TechnicalException;
import de.ozgcloud.eingang.fim.common.xml.DocumentHelper;
import lombok.extern.log4j.Log4j2;

@Log4j2
class FimSchemeCatalogue extends LinkedHashMap<FimSchemeIdentifier, FimScheme> { // NOSONAR

	private static final long serialVersionUID = 1L;

	private final transient FimProperties fimProperties;
	private final transient ResourceLoader resourceLoader;
	private final transient FimSchemeAdapterCatalogue adapterCatalogue;

	private static final FimSchemeAdapter DEFAULT_FIM_SCHEME_ADAPTER = new FimSchemeAdapter() {
		@Override
		public FimSchemeIdentifier getIdentifier() {
			return null;
		}
	};

	public FimSchemeCatalogue(FimProperties fimProperties, ResourceLoader resourceLoader, FimSchemeAdapterCatalogue adapterCatalogue) {
		this.fimProperties = fimProperties;
		this.resourceLoader = resourceLoader;
		this.adapterCatalogue = adapterCatalogue;

		initSchemes();
	}

	private void initSchemes() {
		try {
			initScheme();

			initUnknownScheme();
		} catch (ParserConfigurationException | IOException | SAXException e) {
			throw new TechnicalException("Error on initialising FimSchemes.", e);
		}
	}

	private void initScheme() throws ParserConfigurationException, IOException, SAXException {
		for (var fimSchemaLocation : fimProperties.getSchemeLocations()) {
			put(loadFimScheme(fimSchemaLocation.trim()));
		}
	}

	private void initUnknownScheme() throws ParserConfigurationException {
		put(createScheme(DocumentHelper.getDocumentBuilder().newDocument(), UnknownSchemeAdapter.UNKNOWN_SCHEME_NAME));
	}

	private FimScheme loadFimScheme(String path) throws ParserConfigurationException, IOException, SAXException {
		LOG.debug(() -> "Load FIM schema: " + path);
		var doc = DocumentHelper.parse(getFile(path));
		return createScheme(doc, getTargetNamespace(doc));
	}

	private void put(FimScheme scheme) {
		this.put(scheme.getIdentifier(), scheme);
	}

	private File getFile(String path) throws IOException {
		return resourceLoader.getResource("classpath:" + path).getFile();
	}

	private String getTargetNamespace(Document doc) {
		return doc.getDocumentElement().getAttribute("targetNamespace");
	}

	private FimScheme createScheme(Document doc, String identifierName) {
		var identifier = getIdentifierFromString(identifierName);
		var adapter = getFimSchemeAdapter(identifier);
		return new FimScheme(doc, identifier, adapter);
	}

	private FimSchemeIdentifier getIdentifierFromString(String name) {
		return FimSchemeIdentifier.fromString(name);
	}

	private FimSchemeAdapter getFimSchemeAdapter(FimSchemeIdentifier identifier) {
		return adapterCatalogue.getOrDefault(identifier, DEFAULT_FIM_SCHEME_ADAPTER);
	}
}