/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.eingang.fim;

import java.util.Objects;
import java.util.Optional;

import org.apache.commons.lang3.StringUtils;
import org.springframework.stereotype.Service;
import org.w3c.dom.Document;

import de.ozgcloud.eingang.common.formdata.FormData;
import de.ozgcloud.eingang.common.formdata.FormData.FormDataControl;
import de.ozgcloud.eingang.common.formdata.FormData.Representations;
import de.ozgcloud.eingang.common.formdata.IncomingFile;
import de.ozgcloud.eingang.fim.common.errorhandling.FimException;
import de.ozgcloud.eingang.fim.common.xml.DocumentHelper;
import lombok.RequiredArgsConstructor;
import lombok.extern.log4j.Log4j2;

@Log4j2
@RequiredArgsConstructor
@Service
public class FimService {

	static final String DEFAULT_FORMDATA_REPRESENTATION_NAME = "Antrag.xml";

	private final FimSchemeHelper schemeHelper;

	public Optional<IncomingFile> findPrimaryRepresentation(FormData formData) {
		var representationName = getRepresentationName(formData.getControl());
		return formData.getRepresentations().stream()
				.filter(file -> file.getName().endsWith(representationName))
				.findFirst();
	}

	String getRepresentationName(FormDataControl formDataControl) {
		return formDataControl.getRepresentations().map(Representations::getPrimaryFormDataRepresentation)
				.orElseGet(() -> {
					LOG.warn("No entry point found in metadata file for fim data mapping. Trying default.");
					return DEFAULT_FORMDATA_REPRESENTATION_NAME;
				});
	}

	public FimScheme getScheme(Document document) {
		var fimSchemaName = getSchemeName(document);
		var fimScheme = schemeHelper.getScheme(fimSchemaName);

		if (Objects.isNull(fimScheme)) {
			LOG.error("Cannot find schema for: {}", fimSchemaName);
			return schemeHelper.getDefaultScheme();
		}
		return fimScheme;
	}

	String getSchemeName(Document doc) {
		return Optional.of(DocumentHelper.extractSchemeName(doc))
				.filter(StringUtils::isNotEmpty)
				.orElseThrow(() -> new FimException("XML Document does not provide a scheme."));
	}
}