/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.eingang.fim;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.io.InputStream;
import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.Optional;

import org.junit.jupiter.api.AfterEach;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.MockedStatic;
import org.mockito.Spy;
import org.w3c.dom.Document;

import de.ozgcloud.common.test.ReflectionTestUtils;
import de.ozgcloud.eingang.common.formdata.FormData;
import de.ozgcloud.eingang.common.formdata.FormDataTestFactory;
import de.ozgcloud.eingang.common.formdata.FormHeaderTestFactory;
import de.ozgcloud.eingang.common.formdata.IncomingFile;
import de.ozgcloud.eingang.common.formdata.IncomingFileTestFactory;
import de.ozgcloud.eingang.fim.FimDocumentHelper.FimDocumentProcessor;
import de.ozgcloud.eingang.fim.common.xml.DocumentHelper;

class FimBasedAdapterTest {

	@Spy
	@InjectMocks
	private FimBasedAdapter adapter;
	@Mock
	private FimService service;

	@DisplayName("Is responsible")
	@Nested
	class TestIsResponsible {

		@Test
		void shouldBeTrueOnFimFormEngine() {
			var formData = FormDataTestFactory.createBuilder()
					.header(FormHeaderTestFactory.createBuilder().formEngineName(FimBasedAdapter.FIM_FORM_ENGINE_NAME).build()).build();

			var isResponsible = adapter.isResponsible(formData);

			assertThat(isResponsible).isTrue();
		}
	}

	@DisplayName("Parse formData")
	@Nested
	class TestParseFormData {

		private final FormData formData = FormDataTestFactory.create();
		private final FormData enrichedFormData = FormDataTestFactory.create();
		private final FormData processedFormData = FormDataTestFactory.create();

		@Mock
		private FimEngineBasedMapper mapper;

		@BeforeEach
		void initMocks() {
			ReflectionTestUtils.setField(adapter, "mappers", List.of(mapper));

			doReturn(enrichedFormData).when(adapter).addFormDataDataMap(any());
			doReturn(processedFormData).when(mapper).parseFormData(any());
		}

		@Test
		void shouldCallAddFormDataDataMap() {
			adapter.parseFormData(formData);

			verify(adapter).addFormDataDataMap(formData);
		}

		@Test
		void shouldCallMapper() {
			adapter.parseFormData(formData);

			verify(mapper).parseFormData(enrichedFormData);
		}

		@Test
		void shouldReturnProcessedData() {
			var parsedFormData = adapter.parseFormData(formData);

			assertThat(parsedFormData).isEqualTo(processedFormData);
		}
	}

	@DisplayName("Add formData dataMap")
	@Nested
	class TestAddFormDataDataMap {

		private final FormData formData = FormDataTestFactory.createBuilder().formData(null).build();
		private final Map<String, Object> parsedFormDataDataMap = Collections.emptyMap();

		@BeforeEach
		void initMocks() {
			doReturn(parsedFormDataDataMap).when(adapter).parseRequestData(any());
		}

		@Test
		void shouldCallParseRequestData() {
			adapter.addFormDataDataMap(formData);

			verify(adapter).parseRequestData(formData);
		}

		@Test
		void shouldHaveAddFormData() {
			var enrichedFormData = adapter.addFormDataDataMap(formData);

			assertThat(enrichedFormData.getFormData()).isEqualTo(parsedFormDataDataMap);
		}
	}

	@DisplayName("Parse requestData")
	@Nested
	class TestParseRequestData {

		private final FormData formData = FormDataTestFactory.create();

		@DisplayName("on existing primary representation")
		@Nested
		class TestOnExistingPrimaryRepresentation {

			private final IncomingFile representation = IncomingFileTestFactory.create();
			private final Map<String, Object> parsedData = Collections.emptyMap();

			@BeforeEach
			void initMocks() {
				when(service.findPrimaryRepresentation(any())).thenReturn(Optional.of(representation));
				doReturn(parsedData).when(adapter).doParse(any());
			}

			@Test
			void shouldCallServiceToFindPrimaryRepresentation() {
				adapter.parseRequestData(formData);

				verify(service).findPrimaryRepresentation(formData);
			}

			@Test
			void shouldCallDoParse() {
				adapter.parseRequestData(formData);

				verify(adapter).doParse(representation);
			}

			@Test
			void shouldReturnParsedData() {
				var parsedRequestData = adapter.parseRequestData(formData);

				assertThat(parsedRequestData).isEqualTo(parsedData);
			}
		}

		@Test
		void shouldReturnEmptyMapOnMissingPrimaryRepresentation() {
			when(service.findPrimaryRepresentation(any())).thenReturn(Optional.empty());

			var parsedRequestData = adapter.parseRequestData(formData);

			assertThat(parsedRequestData).isEmpty();
		}
	}

	@DisplayName("Do parse")
	@Nested
	class TestDoParse {

		@Mock
		private IncomingFile file;

		private final InputStream inputStream = InputStream.nullInputStream();

		@BeforeEach
		void initMocks() {
			when(file.getContentStream()).thenReturn(inputStream);
		}

		@Test
		void shouldCallDocumentHelperParse() {
			try (var helper = mockStatic(DocumentHelper.class)) {
				adapter.doParse(file);

				helper.verify(() -> DocumentHelper.parse(inputStream));
			}
		}

		@Test
		void shouldCallProcessData() {
			try (var helper = mockStatic(DocumentHelper.class)) {
				var parsedContent = mock(Document.class);
				helper.when(() -> DocumentHelper.parse(any(InputStream.class))).thenReturn(parsedContent);

				adapter.doParse(file);

				verify(adapter).processData(parsedContent);
			}
		}

		@Test
		void shouldReturnProcessedContent() {
			try (var helper = mockStatic(DocumentHelper.class)) {
				var parsedContent = mock(Document.class);
				helper.when(() -> DocumentHelper.parse(any(InputStream.class))).thenReturn(parsedContent);
				var processedMap = Collections.emptyMap();
				doReturn(processedMap).when(adapter).processData(any());

				var parsedData = adapter.doParse(file);

				assertThat(parsedData).isEqualTo(processedMap);
			}
		}

		@Test
		void shouldReturnEmptyMapOnException() {
			try (var helper = mockStatic(DocumentHelper.class)) {
				helper.when(() -> DocumentHelper.parse(any(InputStream.class))).thenThrow(new RuntimeException());

				var parsedData = adapter.doParse(file);

				assertThat(parsedData).isEmpty();
			}
		}
	}

	@DisplayName("Process data")
	@Nested
	class TestProcessData {

		@Mock
		private Document content;
		@Mock
		private FimScheme scheme;

		@Mock
		private FimDocumentProcessor processor;

		private MockedStatic<FimDocumentHelper> helper;
		private final Map<String, Object> dataMap = Collections.<String, Object>emptyMap();

		@BeforeEach
		void initMocks() {
			helper = mockStatic(FimDocumentHelper.class);
			helper.when(() -> FimDocumentHelper.createProcessor(any(), any())).thenReturn(processor);

			when(processor.process()).thenReturn(dataMap);
			when(service.getScheme(any())).thenReturn(scheme);
		}

		@AfterEach
		void tearDown() {
			helper.close();
		}

		@Test
		void shouldCallServiceToGetScheme() {
			adapter.processData(content);

			verify(service).getScheme(content);
		}

		@Test
		void shouldCreateFimDocumentHelperProcessor() {
			adapter.processData(content);

			helper.verify(() -> FimDocumentHelper.createProcessor(content, scheme));
		}

		@Test
		void shouldProcess() {
			adapter.processData(content);

			verify(processor).process();
		}

		@Test
		void shouldReturnProcessedData() {
			var processedData = adapter.processData(content);

			assertThat(processedData).isEqualTo(dataMap);
		}
	}
}
