/*
 * Copyright (C) 2025 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.eingang.fim;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.util.List;
import java.util.Optional;

import org.apache.commons.lang3.StringUtils;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;
import org.w3c.dom.Document;

import com.thedeanda.lorem.LoremIpsum;

import de.ozgcloud.eingang.common.formdata.FormData;
import de.ozgcloud.eingang.common.formdata.FormData.Representations;
import de.ozgcloud.eingang.common.formdata.FormDataControlTestFactory;
import de.ozgcloud.eingang.common.formdata.FormDataTestFactory;
import de.ozgcloud.eingang.common.formdata.IncomingFile;
import de.ozgcloud.eingang.common.formdata.IncomingFileTestFactory;
import de.ozgcloud.eingang.fim.common.errorhandling.FimException;
import de.ozgcloud.eingang.fim.common.xml.DocumentHelper;

class FimServiceTest {

	@Spy
	@InjectMocks
	private FimService service;
	@Mock
	private FimSchemeHelper schemeHelper;

	@DisplayName("Find primary representation")
	@Nested
	class TestFindPrimaryRepresentation {

		private final String representationName = LoremIpsum.getInstance().getName();
		private final IncomingFile representationFile = IncomingFileTestFactory.createBuilder()
				.name(representationName)
				.build();
		private final IncomingFile file = IncomingFileTestFactory.create();
		private final FormData formData = FormDataTestFactory.createBuilder()
				.clearRepresentations().representations(List.of(representationFile, file))
				.build();

		@BeforeEach
		void initMocks() {
			doReturn(representationName).when(service).getRepresentationName(any());
		}

		@Test
		void shouldCallGetRepresentationName() {
			service.findPrimaryRepresentation(formData);

			verify(service).getRepresentationName(FormDataTestFactory.FORM_DATA_CONTROL);
		}

		@Test
		void shouldReturnPrimaryRepresentationIfExists() {
			var primaryRepresentation = service.findPrimaryRepresentation(formData);

			assertThat(primaryRepresentation).hasValue(representationFile);
		}

		@Test
		void shouldReturnEmptyIfMissing() {
			var formDataWithoutMatchingRepresentation = FormDataTestFactory.createBuilder().clearRepresentations().representation(file).build();

			var primaryRepresentation = service.findPrimaryRepresentation(formDataWithoutMatchingRepresentation);

			assertThat(primaryRepresentation).isEmpty();
		}
	}

	@DisplayName("Get representation name")
	@Nested
	class TestGetRepresentationName {

		@Test
		void shouldReturnPrimaryRepresentationName() {
			var result = service.getRepresentationName(FormDataControlTestFactory.create());

			assertThat(result).isEqualTo(FormDataControlTestFactory.PRIMARY_FORM_DATA_REPRESENTATION);
		}

		@Test
		void shouldReturnDefaultNameOnMissingRepresentations() {
			var formDataControl = FormDataControlTestFactory.createBuilder().representations(Optional.empty()).build();

			var result = service.getRepresentationName(formDataControl);

			assertThat(result).isEqualTo(FimService.DEFAULT_FORMDATA_REPRESENTATION_NAME);
		}

		@Test
		void shouldReturnDefaultOnMissingPrimaryRepresentation() {
			var formDataControl = FormDataControlTestFactory.createBuilder()
					.representations(Optional.of(Representations.builder().build()))
					.build();

			var result = service.getRepresentationName(formDataControl);

			assertThat(result).isEqualTo(FimService.DEFAULT_FORMDATA_REPRESENTATION_NAME);
		}
	}

	@DisplayName("Get scheme")
	@Nested
	class TestGetScheme {

		@Mock
		private Document doc;
		@Mock
		private FimScheme fimScheme;

		private final String schemeName = LoremIpsum.getInstance().getName();

		@BeforeEach
		void initMocks() {
			doReturn(schemeName).when(service).getSchemeName(any());
		}

		@Test
		void shouldCallGetSchemeName() {
			service.getScheme(doc);

			verify(service).getScheme(doc);
		}

		@DisplayName("on existing scheme")
		@Nested
		class TestOnExistingScheme {

			@BeforeEach
			void mock() {
				when(schemeHelper.getScheme(any())).thenReturn(fimScheme);
			}

			@Test
			void shouldCallSchemeHelper() {
				service.getScheme(doc);

				verify(schemeHelper).getScheme(schemeName);
			}

			@Test
			void shouldReturnScheme() {
				var scheme = service.getScheme(doc);

				assertThat(scheme).isEqualTo(fimScheme);
			}
		}

		@DisplayName("on missing scheme")
		@Nested
		class TestOnMissingScheme {

			@BeforeEach
			void mock() {
				when(schemeHelper.getScheme(any())).thenReturn(null);
				when(schemeHelper.getDefaultScheme()).thenReturn(fimScheme);
			}

			@Test
			void shouldCallSchemeHelperToGetDefaultScheme() {
				service.getScheme(doc);

				verify(schemeHelper).getDefaultScheme();
			}

			@Test
			void shouldReturnDefaultScheme() {
				var scheme = service.getScheme(doc);

				assertThat(scheme).isEqualTo(fimScheme);
			}
		}
	}

	@DisplayName("Get scheme name")
	@Nested
	class TestGetSchemeName {

		@Mock
		private Document doc;

		private final String schemeName = LoremIpsum.getInstance().getName();

		@Test
		void shouldCallExtractSchemeName() {
			try (var helper = mockStatic(DocumentHelper.class)) {
				helper.when(() -> DocumentHelper.extractSchemeName(doc)).thenReturn(schemeName);

				service.getSchemeName(doc);

				helper.verify(() -> DocumentHelper.extractSchemeName(doc));
			}
		}

		@Test
		void shouldThrowExceptionOnEmptyName() {
			try (var helper = mockStatic(DocumentHelper.class)) {
				helper.when(() -> DocumentHelper.extractSchemeName(doc)).thenReturn(StringUtils.EMPTY);

				assertThatThrownBy(() -> service.getSchemeName(doc)).isInstanceOf(FimException.class);
			}
		}

		@Test
		void shouldReturnValue() {
			try (var helper = mockStatic(DocumentHelper.class)) {
				helper.when(() -> DocumentHelper.extractSchemeName(doc)).thenReturn(schemeName);

				var name = service.getSchemeName(doc);

				assertThat(name).isEqualTo(schemeName);
			}
		}
	}
}