/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.eingang.formcycle;

import java.io.IOException;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.Map;
import java.util.Objects;
import java.util.Optional;
import java.util.UUID;

import org.springframework.stereotype.Controller;
import org.springframework.web.bind.annotation.PostMapping;
import org.springframework.web.bind.annotation.RequestMapping;
import org.springframework.web.bind.annotation.RequestPart;
import org.springframework.web.bind.annotation.ResponseBody;
import org.springframework.web.multipart.MultipartFile;

import de.ozgcloud.common.binaryfile.TempFileUtils;
import de.ozgcloud.common.errorhandling.TechnicalException;
import de.ozgcloud.eingang.common.formdata.FormData;
import de.ozgcloud.eingang.common.formdata.IncomingFile;
import de.ozgcloud.eingang.common.formdata.IncomingFileGroup;
import de.ozgcloud.eingang.common.vorgang.VorgangNummerSupplier;
import de.ozgcloud.eingang.semantik.SemantikAdapter;
import de.ozgcloud.eingang.semantik.enginebased.FilesMapperHelper;
import lombok.RequiredArgsConstructor;
import lombok.extern.log4j.Log4j2;

@Log4j2
@Controller
@ResponseBody
@RequestMapping("formData")
@RequiredArgsConstructor
class FormDataController {

	public static final String HTTP_TYPE_PROTOBUF = "application/x-protobuf";

	private final FormCycleFormDataMapper mapper;
	private final SemantikAdapter semantikAdapter;
	private final VorgangNummerSupplier vorgangNummerSupplier;
	private final FormDataHtmlCleaner formDataHtmlCleaner;
	private final FormCycleServiceKontoMapper serviceKontoMapper;

	@PostMapping(consumes = "multipart/form-data", produces = HTTP_TYPE_PROTOBUF)
	public FormCycleConfirmationResponse receiveFormData(@RequestPart FormCycleFormData formData,
			@RequestPart(required = false) Optional<Collection<MultipartFile>> representations,
			@RequestPart(required = false) Optional<Collection<MultipartFile>> attachments) {

		FormData mappedFormData = mapper.toFormData(formData);
		mappedFormData = formDataHtmlCleaner.clean(mappedFormData);
		mappedFormData = addRepresentations(representations, mappedFormData);
		mappedFormData = addFiles(formData, attachments, mappedFormData);
		mappedFormData = addServiceKonto(formData, mappedFormData);
		mappedFormData = addVorgangNummer(mappedFormData);

		semantikAdapter.processFormData(mappedFormData);

		return FormCycleConfirmationResponse.newBuilder().setVorgangNummer(mappedFormData.getHeader().getVorgangNummer()).build();
	}

	private FormData addVorgangNummer(FormData mappedFormData) {
		var formDataHeader = mappedFormData.getHeader().toBuilder().vorgangNummer(vorgangNummerSupplier.get()).build();
		return mappedFormData.toBuilder().header(formDataHeader).build();
	}

	private FormData addRepresentations(Optional<Collection<MultipartFile>> files, FormData mappedFormData) {
		Collection<IncomingFile> representations = buildIncomingFiles(files);
		return mappedFormData.toBuilder().representations(representations).numberOfRepresentations(representations.size()).build();
	}

	private Collection<IncomingFile> buildIncomingFiles(Optional<Collection<MultipartFile>> files) {
		return files.orElseGet(Collections::emptyList).stream().map(this::buildIncomingFile).toList();
	}

	FormData addFiles(FormCycleFormData inFormData, Optional<Collection<MultipartFile>> attachments, FormData mappedFormData) {
		var groups = new AttachmentGroupsBuilder(inFormData.getAttachmentGroupList(), attachments).buildGroups();

		return mappedFormData.toBuilder().attachments(groups).numberOfAttachments(FilesMapperHelper.countAttachedFiles(groups)).build();
	}

	private IncomingFile buildIncomingFile(MultipartFile multipartFile) {
		try {
			return IncomingFile.builder()
					.id(UUID.randomUUID().toString())
					.name(multipartFile.getOriginalFilename())
					.size(multipartFile.getSize())
					.contentType(multipartFile.getContentType())
					.file(TempFileUtils.writeTmpFile(multipartFile.getInputStream()))
					.build();
		} catch (IOException e) {
			throw new TechnicalException("Error reading incoming file", e);
		}
	}

	class AttachmentGroupsBuilder {

		private final Collection<FormCycleAttachmentGroup> attachmentGroups;
		private final Map<String, IncomingFile> nameToFile = new HashMap<>();

		public AttachmentGroupsBuilder(Collection<FormCycleAttachmentGroup> attachmentGroups, Optional<Collection<MultipartFile>> attachmentFiles) {
			this.attachmentGroups = attachmentGroups;

			attachmentFiles.ifPresent(files -> files.stream().map(FormDataController.this::buildIncomingFile)
					.forEach(file -> nameToFile.put(file.getName(), file)));
		}

		public Collection<IncomingFileGroup> buildGroups() {
			return attachmentGroups.stream().map(this::buildGroup).toList();
		}

		IncomingFileGroup buildGroup(FormCycleAttachmentGroup group) {
			var groupBuilder = IncomingFileGroup.builder().name(group.getName());

			group.getFileIdList().stream().map(this::getFile)
					.filter(Optional::isPresent).map(Optional::get)
					.forEach(groupBuilder::file);

			return groupBuilder.build();
		}

		Optional<IncomingFile> getFile(String name) {
			var file = nameToFile.get(name);

			if (Objects.isNull(file)) {
				LOG.warn("Cannot find Attachment-File with name '{}'.", name);
				return Optional.empty();
			}
			return Optional.of(file);
		}
	}

	FormData addServiceKonto(FormCycleFormData formData, FormData mappedFormData) {
		if (formData.hasServiceKonto()) {
			mappedFormData.getHeader().setServiceKonto(serviceKontoMapper.fromGrpc(formData.getServiceKonto()));
		}
		return mappedFormData;
	}
}