/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.eingang.formcycle;

import static de.ozgcloud.eingang.common.formdata.IncomingFileTestFactory.*;
import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;
import static org.springframework.test.web.servlet.request.MockMvcRequestBuilders.*;
import static org.springframework.test.web.servlet.result.MockMvcResultMatchers.*;

import java.io.ByteArrayOutputStream;
import java.util.Collections;
import java.util.Optional;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.ArgumentCaptor;
import org.mockito.Captor;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;
import org.springframework.http.converter.protobuf.ProtobufHttpMessageConverter;
import org.springframework.mock.web.MockMultipartFile;
import org.springframework.test.web.servlet.MockMvc;
import org.springframework.test.web.servlet.ResultActions;
import org.springframework.test.web.servlet.setup.MockMvcBuilders;

import de.ozgcloud.eingang.common.formdata.FormData;
import de.ozgcloud.eingang.common.formdata.FormDataTestFactory;
import de.ozgcloud.eingang.common.formdata.FormHeader;
import de.ozgcloud.eingang.common.formdata.FormHeaderTestFactory;
import de.ozgcloud.eingang.common.formdata.IncomingFileGroupTestFactory;
import de.ozgcloud.eingang.common.formdata.IncomingFileTestFactory;
import de.ozgcloud.eingang.common.formdata.ServiceKonto;
import de.ozgcloud.eingang.common.formdata.ServiceKontoTestFactory;
import de.ozgcloud.eingang.common.vorgang.VorgangNummerSupplier;
import de.ozgcloud.eingang.semantik.SemantikAdapter;
import lombok.SneakyThrows;

class FormDataControllerTest {

	@Spy
	@InjectMocks
	private FormDataController controller;

	@Mock
	private FormCycleFormDataMapper mapper;
	@Mock
	private SemantikAdapter semantikAdapter;
	@Mock
	private VorgangNummerSupplier vorgangNummerSupplier;
	@Mock
	private FormDataHtmlCleaner htmlCleaner;
	@Mock
	private FormCycleServiceKontoMapper serviceKontoMapper;

	private MockMvc mockMvc;

	@BeforeEach
	void init() {
		mockMvc = MockMvcBuilders.standaloneSetup(controller)
				.setMessageConverters(new ProtobufHttpMessageConverter())
				.build();
	}

	@Nested
	class ReceiveFormData {

		private FormData mappedFormData = FormDataTestFactory.create();

		@Captor
		private ArgumentCaptor<FormData> formDataCaptor;

		@BeforeEach
		void init() {
			when(htmlCleaner.clean(any())).thenReturn(mappedFormData);
			when(mapper.toFormData(any())).thenReturn(mappedFormData);
			when(vorgangNummerSupplier.get()).thenReturn(FormHeaderTestFactory.VORGANG_NUMMER);
		}

		@Test
		void shouldReturnSuccess() throws Exception {
			doPostRequest().andExpect(status().is2xxSuccessful());
		}

		@Test
		void shouldCallVorgangNummerSupplier() {
			doPostRequest();

			verify(vorgangNummerSupplier).get();
		}

		@Test
		@SneakyThrows
		void shouldResponseWithVorgangNummer() {

			var confirmation = FormCycleConfirmationResponse.parseFrom(
					doPostRequest().andReturn().getResponse().getContentAsByteArray());

			assertThat(confirmation.getVorgangNummer()).isEqualTo(FormHeaderTestFactory.VORGANG_NUMMER);
		}

		@Test
		void shouldSetVorgangNummer() {
			doPostRequest();

			verify(semantikAdapter).processFormData(formDataCaptor.capture());
			assertThat(formDataCaptor.getValue().getHeader().getRequestId()).isEqualTo(FormHeaderTestFactory.REQUEST_ID);
		}

		@Test
		void shouldCallHtmlCleaner() {
			doPostRequest();

			verify(htmlCleaner).clean(mappedFormData);
		}

		@Test
		void shouldCallMapper() {
			doPostRequest();

			verify(mapper).toFormData(notNull());
		}

		@Test
		void shouldCallSemantikAdapter() {
			doPostRequest();

			verify(semantikAdapter).processFormData(formDataCaptor.capture());
			assertThat(formDataCaptor.getValue()).usingRecursiveComparison()
					.ignoringFields("representations", "attachments", "numberOfAttachments")
					.isEqualTo(mappedFormData);
		}

		@Test
		void shouldCallAddServiceKonto() {
			doReturn(mappedFormData).when(controller).addFiles(any(), any(), any());

			doPostRequest();

			verify(controller).addServiceKonto(notNull(), eq(mappedFormData));
		}

		@Nested
		class Representations {

			@BeforeEach
			void init() {
				var formData = FormDataTestFactory.createBuilder().clearRepresentations().numberOfRepresentations(0).build();
				when(mapper.toFormData(any())).thenReturn(formData);
				when(htmlCleaner.clean(any())).thenReturn(formData);
			}

			@Test
			void shouldBePresent() {
				doPostRequest();

				verify(semantikAdapter).processFormData(formDataCaptor.capture());
				var formData = formDataCaptor.getValue();
				assertThat(formData.getRepresentations()).hasSize(1);
				assertThat(formData.getNumberOfRepresentations()).isEqualTo(1);
			}

			@Test
			void shouldBeFilled() {
				doPostRequest();

				verify(semantikAdapter).processFormData(formDataCaptor.capture());
				assertThat(formDataCaptor.getValue().getRepresentations()).first().usingRecursiveComparison()
						.ignoringFields("id", "vendorId", "file")
						.isEqualTo(IncomingFileTestFactory.create());
			}

			@Test
			@SneakyThrows
			void shouldBeFineWithoutRepresentations() {
				mockMvc.perform(
						multipart("/formData")
								.file(new MockMultipartFile("formData", null, FormDataController.HTTP_TYPE_PROTOBUF, buildTestFormData())))
						.andExpect(status().isOk());
			}
		}

		@Nested
		class Attachments {
			@BeforeEach
			void init() {
				var formData = FormDataTestFactory.createBuilder().clearAttachments().numberOfAttachments(0).build();
				when(mapper.toFormData(any())).thenReturn(formData);
				when(htmlCleaner.clean(any())).thenReturn(formData);
			}

			@Test
			void shouldHaveGroup() {
				doPostRequest();

				verify(semantikAdapter).processFormData(formDataCaptor.capture());
				assertThat(formDataCaptor.getValue().getAttachments()).hasSize(1);
			}

			@Test
			void shouldSetNumberOfAttachments() {
				doPostRequest();

				verify(semantikAdapter).processFormData(formDataCaptor.capture());
				assertThat(formDataCaptor.getValue().getNumberOfAttachments()).isEqualTo(1);
			}

		}

		@SneakyThrows
		private ResultActions doPostRequest() {
			return mockMvc.perform(
					multipart("/formData")
							.file(new MockMultipartFile("formData", null, FormDataController.HTTP_TYPE_PROTOBUF, buildTestFormData()))
							.file(IncomingFileTestFactory.asMultipartFile("representations"))
							.file(asMultipartFile("attachments",
									createBuilder().name(IncomingFileTestFactory.VENDOR_ID).build())));
		}
	}

	@SneakyThrows
	static byte[] buildTestFormData() {
		return buildTestFormData(FormCycleFormDataTestFactory.create());
	}

	@SneakyThrows
	static byte[] buildTestFormData(FormCycleFormData formData) {
		var out = new ByteArrayOutputStream();
		formData.writeTo(out);
		return out.toByteArray();
	}

	@Nested
	class TestAttachmentGroupsBuilder {

		private FormDataController.AttachmentGroupsBuilder groupsBuilder;

		@BeforeEach
		void init() {
			groupsBuilder = controller.new AttachmentGroupsBuilder(
					Collections.singleton(FormCycleAttachmentGroupTestFactory.create()),
					Optional.of(Collections.singleton(IncomingFileTestFactory.asMultipartFile(NAME))));
		}

		@Nested
		class TestBuildGroup {
			@Test
			void shouldSetName() {
				var group = groupsBuilder.buildGroup(FormCycleAttachmentGroupTestFactory.create());

				assertThat(group.getName()).isEqualTo(IncomingFileGroupTestFactory.NAME);
			}

			@Test
			void shouldHaveFile() {
				var group = groupsBuilder.buildGroup(FormCycleAttachmentGroupTestFactory.create());

				assertThat(group.getFiles()).hasSize(1).first()
						.usingRecursiveComparison().ignoringFields("id", "vendorId", "file")
						.isEqualTo(IncomingFileTestFactory.create());
			}

			@Test
			void shouldBeFineWithoutFile() {
				var group = groupsBuilder.buildGroup(FormCycleAttachmentGroupTestFactory.createBuilder().clearFileId().build());

				assertThat(group.getFiles()).isEmpty();
			}

			@Test
			void shouldBeFineWithMissingFile() {
				var group = groupsBuilder.buildGroup(FormCycleAttachmentGroupTestFactory.createBuilder().clearFileId().addFileId("missing").build());

				assertThat(group.getFiles()).isEmpty();
			}
		}
	}

	@DisplayName("Add serviceKonto")
	@Nested
	class TestServiceKontoMapping {

		@DisplayName("on existing")
		@Nested
		class TestOnExisting {

			private final ServiceKonto serviceKonto = ServiceKontoTestFactory.create();

			@BeforeEach
			void mock() {
				when(serviceKontoMapper.fromGrpc(any(FormCycleServiceKonto.class))).thenReturn(serviceKonto);
			}

			@Test
			void shouldCallMapper() {
				controller.addServiceKonto(FormCycleFormDataTestFactory.create(), buildEmptyFormDataWithHeader());

				verify(serviceKontoMapper).fromGrpc(FormCycleFormDataTestFactory.SERVICE_KONTO);
			}

			@Test
			void shouldMapServiceKontoType() {
				var formData = controller.addServiceKonto(FormCycleFormDataTestFactory.create(), buildEmptyFormDataWithHeader());

				assertThat(getServiceKontoType(formData)).isEqualTo(ServiceKontoTestFactory.TYPE);
			}
		}

		@Test
		void shouldBeNullIfNotExists() {
			var formData = controller.addServiceKonto(FormCycleFormData.newBuilder().build(), buildEmptyFormDataWithHeader());

			assertThat(formData.getHeader().getServiceKonto()).isNull();
		}

		private String getServiceKontoType(FormData formData) {
			return formData.getHeader().getServiceKonto().getType();
		}

		private FormData buildEmptyFormDataWithHeader() {
			return FormData.builder().header(FormHeader.builder().build()).build();
		}
	}
}