/*
 * Copyright (C) 2025 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.eingang.forwarder;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.PipedInputStream;
import java.io.PipedOutputStream;
import java.util.Optional;
import java.util.concurrent.CompletableFuture;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.TimeoutException;
import java.util.function.Function;

import org.apache.commons.io.IOUtils;

import de.ozgcloud.common.errorhandling.TechnicalException;
import de.ozgcloud.eingang.common.formdata.IncomingFile;
import de.ozgcloud.eingang.forwarding.GrpcFileContent;
import lombok.Builder;
import lombok.extern.log4j.Log4j2;

@Log4j2
class ContentCollector {
	private static final int CHUNK_SIZE = 1024 * 64;
	private static final long TIMEOUT_MINUTES = 10;

	private final IncomingFile incomingFile;

	private final PipedOutputStream pipedOutput;
	private final PipedInputStream pipedInput;
	private final CompletableFuture<File> receivingFileContent;

	@Builder
	private ContentCollector(Function<InputStream, CompletableFuture<File>> fileSaver, IncomingFile incomingFile) {
		this.incomingFile = incomingFile;
		try {
			pipedInput = new PipedInputStream(CHUNK_SIZE);
			pipedOutput = new PipedOutputStream(pipedInput);
			receivingFileContent = fileSaver.apply(pipedInput);
		} catch (IOException e) {
			throw new TechnicalException("Upload initialization failed", e);
		}
	}

	public Optional<IncomingFile> collect(GrpcFileContent fileContent) {
		if (fileContent.getIsEndOfFile()) {
			return Optional.of(handleEndOfFile());
		}
		try {
			pipedOutput.write(fileContent.getContent().toByteArray());
		} catch (IOException e) {
			throw new TechnicalException("Error when writing file content.", e);
		}
		return Optional.empty();
	}

	IncomingFile handleEndOfFile() {
		closeOutputPipe();
		return incomingFile.toBuilder().file(getSavedFileContent()).build();
	}

	File getSavedFileContent() {
		try {
			return receivingFileContent.get(TIMEOUT_MINUTES, TimeUnit.MINUTES);
		} catch (ExecutionException | TimeoutException e) {
			throw new TechnicalException("Receiving file failed.", e);
		} catch (InterruptedException e) {
			Thread.currentThread().interrupt();
			throw new TechnicalException("Upload was interrupted.", e);
		} finally {
			closeInputPipe();
		}
	}

	public void close() {
		closeOutputPipe();
		closeInputPipe();
	}

	void closeOutputPipe() {
		IOUtils.closeQuietly(pipedOutput, e -> LOG.error("Cannot close output stream.", e));
	}

	void closeInputPipe() {
		IOUtils.closeQuietly(pipedInput, e -> LOG.error("Cannot close input stream.", e));
	}

}
