/*
 * Copyright (C) 2025 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.eingang.forwarder;

import java.io.File;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.concurrent.CompletableFuture;
import java.util.function.Consumer;
import java.util.function.Function;

import de.ozgcloud.eingang.common.formdata.FormData;
import de.ozgcloud.eingang.common.formdata.IncomingFile;
import de.ozgcloud.eingang.forwarding.GrpcAttachment;
import de.ozgcloud.eingang.forwarding.GrpcRepresentation;
import de.ozgcloud.eingang.forwarding.GrpcRouteForwarding;
import de.ozgcloud.eingang.forwarding.GrpcRouteForwardingRequest;
import de.ozgcloud.eingang.forwarding.GrpcRouteForwardingResponse;
import io.grpc.stub.StreamObserver;
import lombok.Builder;
import lombok.extern.log4j.Log4j2;

@Log4j2
class EingangStubReceiverStreamObserver implements StreamObserver<GrpcRouteForwardingRequest> {

	private final RouteForwardingMapper routeForwardingMapper;
	private final IncomingFileMapper incomingFileMapper;
	private final IncomingFileGroupMapper incomingFileGroupMapper;
	private final Function<InputStream, CompletableFuture<File>> fileSaver;
	private final Consumer<FormData> formDataConsumer;
	private final Consumer<GrpcRouteForwardingResponse> responseConsumer;

	@Builder
	public EingangStubReceiverStreamObserver(RouteForwardingMapper routeForwardingMapper, IncomingFileMapper incomingFileMapper,
			IncomingFileGroupMapper incomingFileGroupMapper, Function<InputStream, CompletableFuture<File>> fileSaver,
			Consumer<FormData> formDataConsumer, Consumer<GrpcRouteForwardingResponse> responseConsumer) {
		this.routeForwardingMapper = routeForwardingMapper;
		this.incomingFileMapper = incomingFileMapper;
		this.incomingFileGroupMapper = incomingFileGroupMapper;
		this.fileSaver = fileSaver;
		this.formDataConsumer = formDataConsumer;
		this.responseConsumer = responseConsumer;
	}

	private FormData formData;
	private final List<IncomingFile> representations = new ArrayList<>();
	private final Map<String, List<IncomingFile>> attachments = new HashMap<>();

	private String groupName;
	private ContentCollector contentCollector;

	@Override
	public synchronized void onNext(GrpcRouteForwardingRequest request) {
		if (request.hasRouteForwarding()) {
			handleRouteForwarding(request.getRouteForwarding());
		}
		if (request.hasAttachment()) {
			handleAttachment(request.getAttachment());
		}
		if (request.hasRepresentation()) {
			handleRepresentation(request.getRepresentation());
		}
	}

	void handleRouteForwarding(GrpcRouteForwarding routeForwarding) {
		if (Objects.nonNull(formData)) {
			throw new IllegalStateException("Received second RouteForwarding. Send only one per request.");
		}
		formData = routeForwardingMapper.toFormData(routeForwarding);
	}

	void handleAttachment(GrpcAttachment attachment) {
		if (attachment.hasFile()) {
			verifyNoOngoingFileDownload();
			contentCollector = buildContentCollector(incomingFileMapper.fromGrpcAttachmentFile(attachment.getFile()));
			groupName = attachment.getFile().getGroupName();
		} else {
			verifyMetadataWasSent();
			contentCollector.collect(attachment.getContent()).ifPresent(this::addAsAttachment);
		}
	}

	void handleRepresentation(GrpcRepresentation representation) {
		if (representation.hasFile()) {
			verifyNoOngoingFileDownload();
			contentCollector = buildContentCollector(incomingFileMapper.fromGrpcRepresentationFile(representation.getFile()));
		} else {
			verifyMetadataWasSent();
			contentCollector.collect(representation.getContent()).ifPresent(this::addAsRepresentation);
		}
	}

	void verifyNoOngoingFileDownload() {
		if (Objects.nonNull(contentCollector)) {
			throw new IllegalStateException("File metadata received before previous file content was completed.");
		}
	}

	private ContentCollector buildContentCollector(IncomingFile incomingFile) {
		return ContentCollector.builder()
				.fileSaver(fileSaver)
				.incomingFile(incomingFile)
				.build();
	}

	void verifyMetadataWasSent() {
		if (Objects.isNull(contentCollector)) {
			throw new IllegalStateException("File content received before metadata.");
		}
	}

	void addAsAttachment(IncomingFile completedIncomingFile) {
		attachments.computeIfAbsent(groupName, s -> new ArrayList<>()).add(completedIncomingFile);
		contentCollector = null;
	}

	void addAsRepresentation(IncomingFile completedIncomingFile) {
		representations.add(completedIncomingFile);
		contentCollector = null;
	}

	@Override
	public synchronized void onError(Throwable t) {
		LOG.error("Error happened. Receiving stream closed.", t);
		contentCollector.close();
	}

	@Override
	public synchronized void onCompleted() {
		formDataConsumer.accept(assembleFormData());
		responseConsumer.accept(GrpcRouteForwardingResponse.getDefaultInstance());
	}

	FormData assembleFormData() {
		if (Objects.isNull(formData)) {
			throw new IllegalStateException("Never received RouteForwarding containing EingangStub and RouteCriteria.");
		}
		return formData.toBuilder()
				.representations(representations)
				.attachments(attachments.entrySet().stream().map(incomingFileGroupMapper::fromMapEntry).toList())
				.numberOfAttachments(attachments.size())
				.numberOfRepresentations(representations.size())
				.build();
	}

}
