/*
 * Copyright (C) 2025 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.eingang.forwarder;

import static org.assertj.core.api.Assertions.*;
import static org.junit.jupiter.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.PipedInputStream;
import java.io.PipedOutputStream;
import java.util.concurrent.CompletableFuture;
import java.util.concurrent.TimeoutException;
import java.util.function.Function;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.Mock;
import org.springframework.test.util.ReflectionTestUtils;

import de.ozgcloud.common.errorhandling.TechnicalException;
import de.ozgcloud.eingang.common.formdata.IncomingFile;
import de.ozgcloud.eingang.common.formdata.IncomingFileTestFactory;
import de.ozgcloud.eingang.forwarding.GrpcFileContent;
import lombok.SneakyThrows;

class ContentCollectorTest {

	private ContentCollector contentCollector;
	@Mock
	private Function<InputStream, CompletableFuture<File>> fileSaver;
	@Mock
	private CompletableFuture<File> fileContentFuture;
	private final IncomingFile incomingFile = IncomingFileTestFactory.createBuilder().file(null).build();

	void setUpContentCollector() {
		when(fileSaver.apply(any())).thenReturn(fileContentFuture);
		contentCollector = spy(ContentCollector.builder().fileSaver(fileSaver).incomingFile(incomingFile).build());
	}

	@Nested
	class TestBuilding {

		private final byte[] content = new byte[] { 1, 2, 3 };

		@Test
		void shouldSetIncomingFile() {
			setUpContentCollector();

			assertThat(getIncomingFile()).isSameAs(incomingFile);
		}

		@Test
		void shouldCreateInputStream() {
			setUpContentCollector();

			assertThat(getPipedInput()).isNotNull();
		}

		@Test
		void shouldCreateOutputStream() {
			setUpContentCollector();

			assertThat(getPipedOutput()).isNotNull();
		}

		@Test
		void shouldCreateConnectedStreams() {
			setUpContentCollector();

			verifyStreamSetUp();
		}

		@SneakyThrows
		private void verifyStreamSetUp() {
			var pipedInput = getPipedInput();
			var pipedOutput = getPipedOutput();
			pipedOutput.write(content);
			pipedOutput.close();
			var readBytes = pipedInput.readAllBytes();
			assertThat(readBytes).isEqualTo(content);
		}

		@Test
		void shouldCallFileSaver() {
			setUpContentCollector();

			verify(fileSaver).apply(getPipedInput());
		}

		@Test
		void shouldSetReceivingFileContent() {
			setUpContentCollector();

			assertThat(getFileContent()).isSameAs(fileContentFuture);
		}
	}

	@Nested
	class TestCollect {

		@Mock
		private PipedOutputStream pipedOutput;

		@BeforeEach
		void setUp() {
			setUpContentCollector();
			setPipedOutput(pipedOutput);
		}

		@Nested
		class TestOnEndOfFile {

			private final GrpcFileContent fileContent = GrpcFileContentTestFactory.createBuilder().setIsEndOfFile(true).build();
			private final IncomingFile completedIncomingFile = IncomingFileTestFactory.create();

			@BeforeEach
			void setUp() {
				doReturn(completedIncomingFile).when(contentCollector).handleEndOfFile();
			}

			@Test
			void shouldCallHandleEndOfFile() {
				contentCollector.collect(fileContent);

				verify(contentCollector).handleEndOfFile();
			}

			@Test
			@SneakyThrows
			void shouldNotWriteContentToOutputStream() {
				contentCollector.collect(fileContent);

				verify(pipedOutput, never()).write(any());
			}

			@Test
			void shouldReturnCompletedIncomingFile() {
				var result = contentCollector.collect(fileContent);

				assertThat(result).contains(completedIncomingFile);
			}
		}

		@Nested
		class TestOnNotEndOfFile {

			private final GrpcFileContent fileContent = GrpcFileContentTestFactory.createBuilder().setIsEndOfFile(false).build();

			@Test
			@SneakyThrows
			void shouldWriteContentToOutputStream() {
				contentCollector.collect(fileContent);

				verify(pipedOutput).write(GrpcFileContentTestFactory.CONTENT);
			}

			@Test
			void shouldNotCallHandleEndOfFile() {
				contentCollector.collect(fileContent);

				verify(contentCollector, never()).handleEndOfFile();
			}

			@Test
			void shouldReturnEmpty() {
				var result = contentCollector.collect(fileContent);

				assertThat(result).isEmpty();
			}

			@Test
			@SneakyThrows
			void shouldThrowTechnicalExceptionOnIOException() {
				doThrow(new IOException()).when(pipedOutput).write(any());

				assertThrows(TechnicalException.class, () -> {
					contentCollector.collect(fileContent);
				});
			}
		}
	}

	@Nested
	class TestHandleEndOfFile {

		private final IncomingFile incomingFile = IncomingFileTestFactory.create();

		@BeforeEach
		void setUp() {
			setUpContentCollector();
			doNothing().when(contentCollector).closeOutputPipe();
			doReturn(incomingFile.getFile()).when(contentCollector).getSavedFileContent();
		}

		@Test
		void shouldCallCloseOutputPipe() {
			contentCollector.handleEndOfFile();

			verify(contentCollector).closeOutputPipe();
		}

		@Test
		void shouldCallGetSavedFileContent() {
			contentCollector.handleEndOfFile();

			verify(contentCollector).getSavedFileContent();
		}

		@Test
		void shouldReturnIncomingFileWithSavedFile() {
			var returnedIncomingFile = contentCollector.handleEndOfFile();

			assertThat(returnedIncomingFile).usingRecursiveComparison().isEqualTo(incomingFile);
		}
	}

	@Nested
	class TestGetSavedFileContent {
		@BeforeEach
		void setUp() {
			setUpContentCollector();
			doNothing().when(contentCollector).closeInputPipe();
		}

		@Nested
		class TestOnNoExceptions {
			@Mock
			private File fileContent;

			@BeforeEach
			@SneakyThrows
			void setUp() {
				when(fileContentFuture.get(anyLong(), any())).thenReturn(fileContent);
			}

			@Test
			void shouldReturnFile() {
				var savedFileContent = contentCollector.getSavedFileContent();

				assertThat(savedFileContent).isSameAs(fileContent);
			}

			@Test
			void shouldCallCloseInputPipe() {
				contentCollector.getSavedFileContent();

				verify(contentCollector).closeInputPipe();
			}
		}

		@Nested
		class TestOnExecutionException {

			@BeforeEach
			void setUp() {
				setFileContent(CompletableFuture.failedFuture(new Exception()));
			}

			@Test
			void shouldThrowTechnicalException() {
				assertThrows(TechnicalException.class, () -> contentCollector.getSavedFileContent());
			}

			@Test
			void shouldCallCloseInputPipe() {
				try {
					contentCollector.getSavedFileContent();
				} catch (TechnicalException e) {
					// expected
				}

				verify(contentCollector).closeInputPipe();
			}
		}

		@Nested
		class TestOnTimeoutException {

			@Mock
			private CompletableFuture<File> fileFuture;

			@BeforeEach
			@SneakyThrows
			void setUp() {
				setFileContent(fileFuture);
				when(fileFuture.get(anyLong(), any())).thenThrow(new TimeoutException());
			}

			@Test
			void shouldThrowTechnicalException() {
				assertThrows(TechnicalException.class, () -> contentCollector.getSavedFileContent());
			}

			@Test
			void shouldCallCloseInputPipe() {
				try {
					contentCollector.getSavedFileContent();
				} catch (TechnicalException e) {
					// expected
				}

				verify(contentCollector).closeInputPipe();
			}
		}

		@Nested
		class TestOnInterruptedException {

			@Mock
			private CompletableFuture<File> fileFuture;

			@BeforeEach
			@SneakyThrows
			void setUp() {
				setFileContent(fileFuture);
				when(fileFuture.get(anyLong(), any())).thenThrow(new InterruptedException());
			}

			@Test
			void shouldThrowTechnicalException() {
				assertThrows(TechnicalException.class, () -> contentCollector.getSavedFileContent());
			}

			@Test
			void shouldInterruptCurrentThread() {
				try {
					contentCollector.getSavedFileContent();
				} catch (TechnicalException e) {
					// expected
				}

				assertThat(Thread.currentThread().isInterrupted()).isTrue();
			}

			@Test
			void shouldCallCloseInputPipe() {
				try {
					contentCollector.getSavedFileContent();
				} catch (TechnicalException e) {
					// expected
				}

				verify(contentCollector).closeInputPipe();
			}
		}
	}

	@Nested
	class TestClose {

		@BeforeEach
		void setUp() {
			setUpContentCollector();
			doNothing().when(contentCollector).closeOutputPipe();
			doNothing().when(contentCollector).closeInputPipe();
		}

		@Test
		void shouldCallCloseOutputPipe() {
			contentCollector.close();

			verify(contentCollector).closeOutputPipe();
		}

		@Test
		void shouldCallCloseInputPipe() {
			contentCollector.close();

			verify(contentCollector).closeInputPipe();
		}
	}

	@Nested
	class TestCloseOutputPipe {

		@Mock
		private PipedOutputStream pipedOutput;

		@BeforeEach
		void setUp() {
			setUpContentCollector();
			setPipedOutput(pipedOutput);
		}

		@Test
		@SneakyThrows
		void shouldClosePipedOutput() {
			contentCollector.closeOutputPipe();

			verify(pipedOutput).close();
		}

		@Test
		@SneakyThrows
		void shouldNotThrowException() {
			doThrow(IOException.class).when(pipedOutput).close();

			assertDoesNotThrow(() -> contentCollector.closeOutputPipe());
		}
	}

	@Nested
	class TestCloseInputPipe {

		@Mock
		private PipedInputStream pipedInput;

		@BeforeEach
		void setUp() {
			setUpContentCollector();
			setPipedInput(pipedInput);
		}

		@Test
		@SneakyThrows
		void shouldClosePipedInput() {
			contentCollector.closeInputPipe();

			verify(pipedInput).close();
		}

		@Test
		@SneakyThrows
		void shouldNotThrowException() {
			doThrow(IOException.class).when(pipedInput).close();

			assertDoesNotThrow(() -> contentCollector.closeInputPipe());
		}
	}

	@SuppressWarnings("unchecked")
	private CompletableFuture<File> getFileContent() {
		return (CompletableFuture<File>) ReflectionTestUtils.getField(contentCollector, "receivingFileContent");
	}

	private void setFileContent(CompletableFuture<File> fileContent) {
		ReflectionTestUtils.setField(contentCollector, "receivingFileContent", fileContent);
	}

	private PipedInputStream getPipedInput() {
		return (PipedInputStream) ReflectionTestUtils.getField(contentCollector, "pipedInput");
	}

	private void setPipedInput(PipedInputStream pipedInput) {
		ReflectionTestUtils.setField(contentCollector, "pipedInput", pipedInput);
	}

	private PipedOutputStream getPipedOutput() {
		return (PipedOutputStream) ReflectionTestUtils.getField(contentCollector, "pipedOutput");
	}

	private void setPipedOutput(PipedOutputStream pipedOutput) {
		ReflectionTestUtils.setField(contentCollector, "pipedOutput", pipedOutput);
	}

	private IncomingFile getIncomingFile() {
		return (IncomingFile) ReflectionTestUtils.getField(contentCollector, "incomingFile");
	}

}
