/*
 * Copyright (C) 2025 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.eingang.forwarder;

import static org.assertj.core.api.Assertions.*;
import static org.junit.jupiter.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.io.File;
import java.io.InputStream;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.concurrent.CompletableFuture;
import java.util.function.Consumer;
import java.util.function.Function;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.Mock;
import org.springframework.test.util.ReflectionTestUtils;

import de.ozgcloud.eingang.common.formdata.FormData;
import de.ozgcloud.eingang.common.formdata.FormDataTestFactory;
import de.ozgcloud.eingang.common.formdata.IncomingFile;
import de.ozgcloud.eingang.common.formdata.IncomingFileGroupTestFactory;
import de.ozgcloud.eingang.common.formdata.IncomingFileTestFactory;
import de.ozgcloud.eingang.forwarding.GrpcAttachment;
import de.ozgcloud.eingang.forwarding.GrpcRepresentation;
import de.ozgcloud.eingang.forwarding.GrpcRouteForwarding;
import de.ozgcloud.eingang.forwarding.GrpcRouteForwardingResponse;

class EingangStubReceiverStreamObserverTest {

	private EingangStubReceiverStreamObserver observer;

	@Mock
	private RouteForwardingMapper routeForwardingMapper;
	@Mock
	private IncomingFileMapper incomingFileMapper;
	@Mock
	private IncomingFileGroupMapper incomingFileGroupMapper;
	@Mock
	private Function<InputStream, CompletableFuture<File>> fileSaver;
	@Mock
	private Consumer<FormData> formDataConsumer;
	@Mock
	private Consumer<GrpcRouteForwardingResponse> responseConsumer;

	@BeforeEach
	void setUp() {
		observer = spy(EingangStubReceiverStreamObserver.builder()
				.fileSaver(fileSaver)
				.routeForwardingMapper(routeForwardingMapper)
				.incomingFileMapper(incomingFileMapper)
				.incomingFileGroupMapper(incomingFileGroupMapper)
				.formDataConsumer(formDataConsumer)
				.responseConsumer(responseConsumer)
				.build());
	}

	@Nested
	class TestOnNext {

		@Nested
		class TestOnRouteForwarding {

			@BeforeEach
			void mock() {
				doNothing().when(observer).handleRouteForwarding(any());
			}

			@Test
			void shouldCallHandleRouteForwarding() {
				observer.onNext(GrpcRouteForwardingRequestTestFactory.createWithRouteForwarding());

				verify(observer).handleRouteForwarding(GrpcRouteForwardingRequestTestFactory.ROUTE_FORWARDING);
			}

			@Test
			void shouldNotCallHandleAttachment() {
				observer.onNext(GrpcRouteForwardingRequestTestFactory.createWithRouteForwarding());

				verify(observer, never()).handleAttachment(any());
			}

			@Test
			void shouldNotCallHandleRepresentation() {
				observer.onNext(GrpcRouteForwardingRequestTestFactory.createWithRouteForwarding());

				verify(observer, never()).handleRepresentation(any());
			}
		}

		@Nested
		class TestOnAttachment {
			@Test
			void shouldCallHandleAttachment() {
				observer.onNext(GrpcRouteForwardingRequestTestFactory.createWithAttachment());

				verify(observer).handleAttachment(GrpcRouteForwardingRequestTestFactory.ATTACHMENT);
			}

			@Test
			void shouldNotCallHandleRouteForwarding() {
				observer.onNext(GrpcRouteForwardingRequestTestFactory.createWithAttachment());

				verify(observer, never()).handleRouteForwarding(any());
			}

			@Test
			void shouldNotCallHandleRepresentation() {
				observer.onNext(GrpcRouteForwardingRequestTestFactory.createWithAttachment());

				verify(observer, never()).handleRepresentation(any());
			}
		}

		@Nested
		class TestOnRepresentation {
			@Test
			void shouldCallHandleRepresentation() {
				observer.onNext(GrpcRouteForwardingRequestTestFactory.createWithRepresentation());

				verify(observer).handleRepresentation(GrpcRouteForwardingRequestTestFactory.REPRESENTATION);
			}

			@Test
			void shouldNotCallHandleRouteForwarding() {
				observer.onNext(GrpcRouteForwardingRequestTestFactory.createWithRepresentation());

				verify(observer, never()).handleRouteForwarding(any());
			}

			@Test
			void shouldNotCallHandleAttachment() {
				observer.onNext(GrpcRouteForwardingRequestTestFactory.createWithRepresentation());

				verify(observer, never()).handleAttachment(any());
			}
		}
	}

	@Nested
	class TestHandleRouteForwarding {

		private final GrpcRouteForwarding routeForwarding = GrpcRouteForwardingTestFactory.create();

		@Test
		void shouldThrowIllegalStateExceptionIfFormDataIsSet() {
			setFormData(FormDataTestFactory.create());

			assertThrows(IllegalStateException.class, () -> observer.handleRouteForwarding(routeForwarding));

		}

		@Test
		void shouldMapRouteForwarding() {
			observer.handleRouteForwarding(routeForwarding);

			verify(routeForwardingMapper).toFormData(routeForwarding);
		}

		@Test
		void shouldSetFormData() {
			var formData = FormDataTestFactory.create();
			when(routeForwardingMapper.toFormData(any())).thenReturn(formData);

			observer.handleRouteForwarding(routeForwarding);

			assertThat(getFormData()).isEqualTo(formData);
		}
	}

	@Nested
	class TestHandleAttachment {

		@Mock
		private ContentCollector contentCollector;

		@Nested
		class TestWithFile {

			private final GrpcAttachment attachmentWithFile = GrpcAttachmentTestFactory.createWithFile();
			private final IncomingFile incomingFile = IncomingFileTestFactory.create();

			@BeforeEach
			void initMock() {
				when(incomingFileMapper.fromGrpcAttachmentFile(any())).thenReturn(incomingFile);
				doNothing().when(observer).verifyNoOngoingFileDownload();
			}

			@Test
			void shouldCallVerifyNoOngoingFileDownload() {
				observer.handleAttachment(attachmentWithFile);

				verify(observer).verifyNoOngoingFileDownload();
			}

			@Test
			void shouldCallIncomingFileMapper() {
				observer.handleAttachment(attachmentWithFile);

				verify(incomingFileMapper).fromGrpcAttachmentFile(GrpcAttachmentTestFactory.FILE);
			}

			@Test
			void shouldSetContentCollector() {
				var expectedContentCollector = ContentCollector.builder().fileSaver(fileSaver).incomingFile(incomingFile).build();

				observer.handleAttachment(attachmentWithFile);

				assertThat(getContentCollector()).usingRecursiveComparison().ignoringFields("pipedInput", "pipedOutput")
						.isEqualTo(expectedContentCollector);
			}

			@Test
			void shouldSetGroupName() {
				observer.handleAttachment(attachmentWithFile);

				assertThat(getGroupName()).isEqualTo(GrpcAttachmentFileTestFactory.GROUP_NAME);
			}

			@Test
			void shouldNotCollectContent() {
				observer.handleAttachment(attachmentWithFile);

				verify(contentCollector, never()).collect(any());
			}
		}

		@Nested
		class TestWithContent {

			private final GrpcAttachment attachmentWithContent = GrpcAttachmentTestFactory.createWithContent();
			private final IncomingFile incomingFile = IncomingFileTestFactory.create();

			@BeforeEach
			void setUp() {
				setContentCollector(contentCollector);
			}

			@Test
			void shouldCallVerifyMetadataWasSent() {
				observer.handleAttachment(attachmentWithContent);

				verify(observer).verifyMetadataWasSent();
			}

			@Test
			void shouldCollectContent() {
				observer.handleAttachment(attachmentWithContent);

				verify(contentCollector).collect(GrpcAttachmentTestFactory.CONTENT);
			}

			@Test
			void shouldNotCallIncomingFileMapper() {
				observer.handleAttachment(attachmentWithContent);

				verify(incomingFileMapper, never()).fromGrpcAttachmentFile(any());
			}

			@Test
			void shouldNotSetGroupName() {
				observer.handleAttachment(attachmentWithContent);

				assertThat(getGroupName()).isNull();
			}

			@Test
			void shouldCallAddAsAttachment() {
				when(contentCollector.collect(any())).thenReturn(Optional.of(incomingFile));

				observer.handleAttachment(attachmentWithContent);

				verify(observer).addAsAttachment(incomingFile);
			}

			@Test
			void shouldNotCallAddAsAttachment() {
				when(contentCollector.collect(any())).thenReturn(Optional.empty());

				observer.handleAttachment(attachmentWithContent);

				verify(observer, never()).addAsAttachment(any());
			}

		}
	}

	@Nested
	class TestHandleRepresentation {

		@Mock
		private ContentCollector contentCollector;

		@Nested
		class TestWithFile {

			private final GrpcRepresentation representationWithFile = GrpcRepresentationTestFactory.createWithFile();
			private final IncomingFile incomingFile = IncomingFileTestFactory.create();

			@BeforeEach
			void mock() {
				when(incomingFileMapper.fromGrpcRepresentationFile(any())).thenReturn(incomingFile);
				doNothing().when(observer).verifyNoOngoingFileDownload();
			}

			@Test
			void shouldCallVerifyNoOngoingFileDownload() {
				observer.handleRepresentation(representationWithFile);

				verify(observer).verifyNoOngoingFileDownload();
			}

			@Test
			void shouldCallIncomingFileMapper() {
				observer.handleRepresentation(representationWithFile);

				verify(incomingFileMapper).fromGrpcRepresentationFile(GrpcRepresentationTestFactory.FILE);
			}

			@Test
			void shouldSetContentCollector() {
				var expectedContentCollector = ContentCollector.builder().fileSaver(fileSaver).incomingFile(incomingFile).build();

				observer.handleRepresentation(representationWithFile);

				assertThat(getContentCollector()).usingRecursiveComparison().ignoringFields("pipedInput", "pipedOutput")
						.isEqualTo(expectedContentCollector);
			}

			@Test
			void shouldNotCollectContent() {
				observer.handleRepresentation(representationWithFile);

				verify(contentCollector, never()).collect(any());
			}
		}

		@Nested
		class TestWithContent {

			private final GrpcRepresentation representationWithContent = GrpcRepresentationTestFactory.createWithContent();
			private final IncomingFile incomingFile = IncomingFileTestFactory.create();

			@BeforeEach
			void setUp() {
				setContentCollector(contentCollector);
			}

			@Test
			void shouldCallVerifyMetadataWasSent() {
				observer.handleRepresentation(representationWithContent);

				verify(observer).verifyMetadataWasSent();
			}

			@Test
			void shouldCollectContent() {
				observer.handleRepresentation(representationWithContent);

				verify(contentCollector).collect(GrpcRepresentationTestFactory.CONTENT);
			}

			@Test
			void shouldNotCallIncomingFileMapper() {
				observer.handleRepresentation(representationWithContent);

				verify(incomingFileMapper, never()).fromGrpcRepresentationFile(any());
			}

			@Test
			void shouldCallAddAsRepresentation() {
				when(contentCollector.collect(any())).thenReturn(Optional.of(incomingFile));

				observer.handleRepresentation(representationWithContent);

				verify(observer).addAsRepresentation(incomingFile);
			}

			@Test
			void shouldNotCallAddAsRepresentation() {
				when(contentCollector.collect(any())).thenReturn(Optional.empty());

				observer.handleRepresentation(representationWithContent);

				verify(observer, never()).addAsRepresentation(any());
			}
		}
	}

	@Nested
	class TestVerifyNoOngoingFileDownload {

		@Test
		void shouldThrowIllegalStateExceptionIfContentCollectorIsNotNull() {
			setContentCollector(mock(ContentCollector.class));

			assertThrows(IllegalStateException.class, () -> observer.verifyNoOngoingFileDownload());
		}

		@Test
		void shouldNotThrowExceptionIfContentCollectorIsNull() {
			assertDoesNotThrow(() -> observer.verifyNoOngoingFileDownload());
		}
	}

	@Nested
	class TestVerifyMetadataWasSent {

		@Test
		void shouldThrowIllegalStateExceptionIfContentCollectorIsNull() {
			assertThrows(IllegalStateException.class, () -> observer.verifyMetadataWasSent());
		}

		@Test
		void shouldNotThrowExceptionIfContentCollectorIsNotNull() {
			setContentCollector(mock(ContentCollector.class));

			assertDoesNotThrow(() -> observer.verifyMetadataWasSent());
		}
	}

	@Nested
	class TestAddAsAttachment {

		@Mock
		private ContentCollector contentCollector;

		private final IncomingFile incomingFile = IncomingFileTestFactory.create();

		@BeforeEach
		void setUp() {
			setContentCollector(contentCollector);
			setGroupName(GrpcAttachmentFileTestFactory.GROUP_NAME);
		}

		@Test
		void shouldAddFileToAttachments() {
			observer.addAsAttachment(incomingFile);

			var attachmentGroup = getAttachments().get(GrpcAttachmentFileTestFactory.GROUP_NAME);
			assertThat(attachmentGroup).usingRecursiveFieldByFieldElementComparator().containsExactly(incomingFile);
		}

		@Test
		void shouldSetContentCollectorToNull() {
			observer.addAsAttachment(incomingFile);

			assertThat(getContentCollector()).isNull();
		}
	}

	@Nested
	class TestAddAsRepresentation {

		@Mock
		private ContentCollector contentCollector;

		private final IncomingFile incomingFile = IncomingFileTestFactory.create();

		@BeforeEach
		void setUp() {
			setContentCollector(contentCollector);
		}

		@Test
		void shouldAddFileToRepresentations() {
			observer.addAsRepresentation(incomingFile);

			assertThat(getRepresentations()).usingRecursiveFieldByFieldElementComparator().containsExactly(incomingFile);
		}

		@Test
		void shouldSetContentCollectorToNull() {
			observer.addAsRepresentation(incomingFile);

			assertThat(getContentCollector()).isNull();
		}
	}

	@Nested
	class TestOnError {

		@Mock
		private ContentCollector contentCollector;

		@BeforeEach
		void mock() {
			setContentCollector(contentCollector);
			setContentCollector(contentCollector);
		}

		@Test
		void shouldCloseContentCollector() {
			observer.onError(new Exception());

			verify(contentCollector).close();
		}
	}

	@Nested
	class TestOnCompleted {

		private final FormData formData = FormDataTestFactory.create();

		@BeforeEach
		void mock() {
			doReturn(formData).when(observer).assembleFormData();
		}

		@Test
		void shouldCallAssembleFormData() {
			observer.onCompleted();

			verify(observer).assembleFormData();
		}

		@Test
		void shouldCallFormDataConsumer() {
			observer.onCompleted();

			verify(formDataConsumer).accept(formData);
		}

		@Test
		void shouldCallResponseConsumer() {
			observer.onCompleted();

			verify(responseConsumer).accept(GrpcRouteForwardingResponse.getDefaultInstance());
		}
	}

	@Nested
	class TestAssembleFormData {

		@Nested
		class TestOnFormDataNotSet {

			@Test
			void shouldThrowIllegalStateException() {
				assertThrows(IllegalStateException.class, () -> observer.assembleFormData());
			}
		}

		@Nested
		class TestOnFormDataSet {

			private final FormData formData = FormDataTestFactory.createBuilder()
					.clearAttachments()
					.clearRepresentations()
					.numberOfAttachments(0)
					.numberOfRepresentations(0)
					.control(null)
					.build();

			private final List<IncomingFile> representations = FormDataTestFactory.REPRESENTATIONS;
			private final Map.Entry<String, List<IncomingFile>> attachmentEntry = Map.entry(
					IncomingFileGroupTestFactory.NAME, IncomingFileGroupTestFactory.INCOMING_FILES);

			@BeforeEach
			void setUp() {
				setFormData(formData);
				setRepresentations(representations);
				setAttachments(Map.ofEntries(attachmentEntry));
			}

			@Test
			void shouldCallIncomingFileGroupMapper() {
				observer.assembleFormData();

				verify(incomingFileGroupMapper).fromMapEntry(attachmentEntry);
			}

			@Test
			void shouldReturnFormData() {
				var expectedFormData = FormDataTestFactory.createBuilder()
						.control(null)
						.build();
				when(incomingFileGroupMapper.fromMapEntry(attachmentEntry)).thenReturn(IncomingFileGroupTestFactory.create());

				var assembledFormData = observer.assembleFormData();

				assertThat(assembledFormData).usingRecursiveComparison().isEqualTo(expectedFormData);
			}
		}
	}

	private FormData getFormData() {
		return (FormData) ReflectionTestUtils.getField(observer, "formData");
	}

	private void setFormData(FormData formData) {
		ReflectionTestUtils.setField(observer, "formData", formData);
	}

	private String getGroupName() {
		return (String) ReflectionTestUtils.getField(observer, "groupName");
	}

	private void setGroupName(String groupName) {
		ReflectionTestUtils.setField(observer, "groupName", groupName);
	}

	@SuppressWarnings("unchecked")
	private List<IncomingFile> getRepresentations() {
		return (List<IncomingFile>) ReflectionTestUtils.getField(observer, "representations");
	}

	private void setRepresentations(List<IncomingFile> representations) {
		ReflectionTestUtils.setField(observer, "representations", representations);
	}

	@SuppressWarnings("unchecked")
	private Map<String, List<IncomingFile>> getAttachments() {
		return (Map<String, List<IncomingFile>>) ReflectionTestUtils.getField(observer, "attachments");
	}

	private void setAttachments(Map<String, List<IncomingFile>> attachments) {
		ReflectionTestUtils.setField(observer, "attachments", attachments);
	}

	private ContentCollector getContentCollector() {
		return (ContentCollector) ReflectionTestUtils.getField(observer, "contentCollector");
	}

	private void setContentCollector(ContentCollector contentCollector) {
		ReflectionTestUtils.setField(observer, "contentCollector", contentCollector);
	}

}
