/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.eingang.forwarder;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.io.ByteArrayInputStream;
import java.io.File;
import java.io.InputStream;
import java.util.concurrent.CompletableFuture;
import java.util.function.Consumer;
import java.util.function.Function;

import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;
import org.springframework.test.util.ReflectionTestUtils;

import com.thedeanda.lorem.LoremIpsum;

import de.ozgcloud.eingang.common.formdata.FormData;
import de.ozgcloud.eingang.common.formdata.FormDataTestFactory;
import de.ozgcloud.eingang.forwarding.GrpcRouteForwardingRequest;
import de.ozgcloud.eingang.forwarding.GrpcRouteForwardingResponse;
import io.grpc.stub.StreamObserver;

class RouteForwardingGrpcServiceTest {

	@InjectMocks
	@Spy
	private RouteForwardingGrpcService service;

	@Mock
	private StreamObserver<GrpcRouteForwardingResponse> responseObserver;

	@Mock
	private RouteForwardingMapper routeForwardingMapper;
	@Mock
	private IncomingFileMapper incomingFileMapper;
	@Mock
	private IncomingFileGroupMapper incomingFileGroupMapper;
	@Mock
	private RouteForwardingService routeForwardingService;
	@Mock
	private FileService fileService;

	@Nested
	class TestRouteForwarding {
		private static final byte[] CONTENT = LoremIpsum.getInstance().getWords(5).getBytes();

		private static final ByteArrayInputStream INPUT_STREAM = new ByteArrayInputStream(CONTENT);

		private final FormData formData = FormDataTestFactory.create();

		@Test
		void shouldSetFileSaver() {
			var observer = service.routeForwarding(responseObserver);

			callFileSaver(observer);
			verify(service).saveFile(INPUT_STREAM);
		}

		@Test
		void shouldSetRouteForwardingMapper() {
			var observer = service.routeForwarding(responseObserver);

			assertThat(getRouteForwardingMapper(observer)).isEqualTo(routeForwardingMapper);
		}

		@Test
		void shouldSetIncomingFileMapper() {
			var observer = service.routeForwarding(responseObserver);

			assertThat(getIncomingFileMapper(observer)).isEqualTo(incomingFileMapper);
		}

		@Test
		void shouldSetIncomingFileGroupMapper() {
			var observer = service.routeForwarding(responseObserver);

			assertThat(getIncomingFileGroupMapper(observer)).isEqualTo(incomingFileGroupMapper);
		}

		@Test
		void shouldSetFormDataConsumer() {
			var observer = service.routeForwarding(responseObserver);

			callFormDataConsumer(observer, formData);
			verify(routeForwardingService).route(formData);
		}

		@Test
		void shouldSetResponseConsumer() {
			var observer = service.routeForwarding(responseObserver);

			callResponseConsumer(observer, GrpcRouteForwardingResponse.getDefaultInstance());
			verify(service).respondWith(responseObserver, GrpcRouteForwardingResponse.getDefaultInstance());
		}

		@SuppressWarnings("unchecked")
		private void callFileSaver(StreamObserver<GrpcRouteForwardingRequest> uploadObserver) {
			var fileSaver = (Function<InputStream, CompletableFuture<File>>) ReflectionTestUtils.getField(uploadObserver, "fileSaver");
			fileSaver.apply(INPUT_STREAM);
		}

		private RouteForwardingMapper getRouteForwardingMapper(StreamObserver<GrpcRouteForwardingRequest> uploadObserver) {
			var routeForwardingMapper = (RouteForwardingMapper) ReflectionTestUtils.getField(uploadObserver, "routeForwardingMapper");
			return routeForwardingMapper;
		}

		private IncomingFileMapper getIncomingFileMapper(StreamObserver<GrpcRouteForwardingRequest> uploadObserver) {
			var incomingFileMapper = (IncomingFileMapper) ReflectionTestUtils.getField(uploadObserver, "incomingFileMapper");
			return incomingFileMapper;
		}

		private IncomingFileGroupMapper getIncomingFileGroupMapper(StreamObserver<GrpcRouteForwardingRequest> uploadObserver) {
			var incomingFileGroupMapper = (IncomingFileGroupMapper) ReflectionTestUtils.getField(uploadObserver, "incomingFileGroupMapper");
			return incomingFileGroupMapper;
		}

		@SuppressWarnings("unchecked")
		private void callFormDataConsumer(StreamObserver<GrpcRouteForwardingRequest> uploadObserver, FormData formData) {
			var formDataConsumer = (Consumer<FormData>) ReflectionTestUtils.getField(uploadObserver, "formDataConsumer");
			formDataConsumer.accept(formData);
		}

		@SuppressWarnings("unchecked")
		private void callResponseConsumer(StreamObserver<GrpcRouteForwardingRequest> uploadObserver, GrpcRouteForwardingResponse response) {
			var responseConsumer = (Consumer<GrpcRouteForwardingResponse>) ReflectionTestUtils.getField(uploadObserver, "responseConsumer");
			responseConsumer.accept(response);
		}
	}

	@Nested
	class TestSaveFile {

		@Mock
		private InputStream inputStream;
		@Mock
		private CompletableFuture<File> fileFuture;

		@Test
		void shouldReturnNull() {
			service.saveFile(inputStream);

			verify(fileService).saveToFile(inputStream);
		}

		@Test
		void shouldReturnFileFuture() {
			when(fileService.saveToFile(any())).thenReturn(fileFuture);

			var result = service.saveFile(inputStream);

			assertThat(result).isSameAs(fileFuture);
		}
	}

	@Nested
	class TestRespondWith {

		private GrpcRouteForwardingResponse response = GrpcRouteForwardingResponse.newBuilder().build();
		@Mock
		private StreamObserver<GrpcRouteForwardingResponse> responseObserver;

		@Test
		void shouldCallOnNext() {
			service.respondWith(responseObserver, response);

			verify(responseObserver).onNext(response);
		}

		@Test
		void shouldCallOnCompletedAfterOnNext() {
			var inOrder = inOrder(responseObserver);

			service.respondWith(responseObserver, response);

			inOrder.verify(responseObserver).onNext(response);
			inOrder.verify(responseObserver).onCompleted();
		}
	}
}
