/*
 * Copyright (C) 2025 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.eingang.forwarder;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mapstruct.factory.Mappers;
import org.mockito.InjectMocks;
import org.mockito.Mock;

import de.ozgcloud.eingang.common.formdata.AntragstellerTestFactory;
import de.ozgcloud.eingang.common.formdata.FormDataTestFactory;
import de.ozgcloud.eingang.common.formdata.ServiceKontoTestFactory;
import de.ozgcloud.eingang.common.formdata.ZustaendigeStelle;
import de.ozgcloud.eingang.forwarding.GrpcRouteForwarding;
import de.ozgcloud.eingang.router.GrpcEingangHeaderTestFactory;
import de.ozgcloud.eingang.router.ServiceKontoMapper;
import de.ozgcloud.vorgang.common.grpc.GrpcFormDataMapper;

class RouteForwardingMapperTest {

	@InjectMocks
	private RouteForwardingMapper mapper = Mappers.getMapper(RouteForwardingMapper.class);

	@Mock
	private GrpcFormDataMapper grpcFormDataMapper;
	@Mock
	private ServiceKontoMapper serviceKontoMapper;

	@Nested
	class TestToFormData {

		private final GrpcRouteForwarding routeForwarding = GrpcRouteForwardingTestFactory.create();

		@BeforeEach
		void mock() {
			when(serviceKontoMapper.fromGrpc(any())).thenReturn(ServiceKontoTestFactory.create());
			when(grpcFormDataMapper.mapFromFormData(GrpcAntragstellerTestFactory.OTHER_DATA)).thenReturn(AntragstellerTestFactory.DATA);
			when(grpcFormDataMapper.mapFromFormData(GrpcEingangStubTestFactory.FORM_DATA)).thenReturn(FormDataTestFactory.FORM_DATA);
		}

		@Test
		void shouldCallServiceKontoMapper() {
			mapper.toFormData(routeForwarding);

			verify(serviceKontoMapper).fromGrpc(GrpcEingangHeaderTestFactory.SERVICE_KONTO);
		}

		@Test
		void shouldCallFormDataMapperWithAntragstellerData() {
			mapper.toFormData(routeForwarding);

			verify(grpcFormDataMapper).mapFromFormData(GrpcAntragstellerTestFactory.OTHER_DATA);
		}

		@Test
		void shouldCallFormDataMapperWithFormData() {
			mapper.toFormData(routeForwarding);

			verify(grpcFormDataMapper).mapFromFormData(GrpcEingangStubTestFactory.FORM_DATA);
		}

		@Test
		void shouldMapHeader() {
			var formDataHeader = mapper.toFormData(routeForwarding).getHeader();

			assertThat(formDataHeader).usingRecursiveComparison().isEqualTo(FormDataTestFactory.HEADER);
		}

		@Test
		void shouldMapAntragsteller() {
			var expectedAntragSteller = AntragstellerTestFactory.createBuilder().firmaName(AntragstellerTestFactory.FIRMA_NAME).build();

			var antragsteller = mapper.toFormData(routeForwarding).getAntragsteller();

			assertThat(antragsteller).usingRecursiveComparison().isEqualTo(expectedAntragSteller);
		}

		@Test
		void shouldMapFormData() {
			var formData = mapper.toFormData(routeForwarding).getFormData();

			assertThat(formData).isSameAs(FormDataTestFactory.FORM_DATA);
		}

		@Test
		void shouldMapZustaendigeStelles() {
			var expectedZustaendigeStelle = ZustaendigeStelle.builder()
					.organisationseinheitenId(GrpcRouteCriteriaTestFactory.ORGANISATION_EINHEIT_ID)
					.build();

			var zustaendigeStelles = mapper.toFormData(routeForwarding).getZustaendigeStelles();

			assertThat(zustaendigeStelles).usingRecursiveFieldByFieldElementComparator().containsExactly(expectedZustaendigeStelle);
		}
	}
}
