/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.eingang.router;

import java.io.InputStream;
import java.util.List;
import java.util.Optional;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.TimeoutException;

import org.apache.commons.io.IOUtils;
import org.springframework.stereotype.Service;

import com.google.protobuf.ByteString;

import de.ozgcloud.common.binaryfile.GrpcFileUploadUtils;
import de.ozgcloud.common.binaryfile.GrpcFileUploadUtils.FileSender;
import de.ozgcloud.common.errorhandling.TechnicalException;
import de.ozgcloud.eingang.common.formdata.FormData;
import de.ozgcloud.eingang.common.formdata.IncomingFile;
import de.ozgcloud.eingang.common.formdata.IncomingFileGroup;
import de.ozgcloud.vorgang.grpc.binaryFile.BinaryFileServiceGrpc.BinaryFileServiceStub;
import de.ozgcloud.vorgang.grpc.binaryFile.GrpcUploadBinaryFileMetaData;
import de.ozgcloud.vorgang.grpc.binaryFile.GrpcUploadBinaryFileRequest;
import de.ozgcloud.vorgang.grpc.binaryFile.GrpcUploadBinaryFileResponse;
import de.ozgcloud.vorgang.grpc.command.GrpcCallContext;
import de.ozgcloud.vorgang.vorgang.GrpcCreateVorgangRequest;
import de.ozgcloud.vorgang.vorgang.GrpcEingang;
import de.ozgcloud.vorgang.vorgang.GrpcFinishCreationRequest;
import de.ozgcloud.vorgang.vorgang.GrpcIncomingFile;
import de.ozgcloud.vorgang.vorgang.GrpcIncomingFileGroup;
import de.ozgcloud.vorgang.vorgang.VorgangServiceGrpc.VorgangServiceBlockingStub;
import io.grpc.stub.CallStreamObserver;
import io.grpc.stub.StreamObserver;
import lombok.AllArgsConstructor;
import lombok.Getter;
import lombok.RequiredArgsConstructor;
import lombok.extern.log4j.Log4j2;

@Log4j2
@AllArgsConstructor
@Service
public class VorgangRemoteService {

	private final VorgangManagerServerResolver vorgangManagerServerResolver;

	private final GrpcEingangMapper grpcEingangMapper;

	public String createVorgang(FormData formData, Optional<String> organisationsEinheitenId) {
		var vorgangServiceStub = getVorgangServiceStub(organisationsEinheitenId);
		var binaryFileServiceStub = getBinaryFileServiceStub(organisationsEinheitenId);

		logConnection(organisationsEinheitenId, vorgangServiceStub.get());

		try {
			return createVorgang(formData, grpcEingangMapper.toEingang(formData, organisationsEinheitenId), vorgangServiceStub.get(), binaryFileServiceStub.get());
		} finally {
			finishStubConnections(List.of(vorgangServiceStub, binaryFileServiceStub));
		}
	}

	void logConnection(Optional<String> organisationsEinheitenId, VorgangServiceBlockingStub vorgangStub) {
		LOG.info("Connecting to vorgang-manager {}; OrganisationsEinheitId: {}.", vorgangStub.getChannel().authority(),
				organisationsEinheitenId);
	}

	private ManagableStub<VorgangServiceBlockingStub> getVorgangServiceStub(Optional<String> organisationsEinheitenId) {
		return vorgangManagerServerResolver.resolveVorgangServiceBlockingStubByOrganisationseinheitenId(organisationsEinheitenId);
	}

	private ManagableStub<BinaryFileServiceStub> getBinaryFileServiceStub(Optional<String> organisationsEinheitenId) {
		return vorgangManagerServerResolver.resolveBinaryFileServiceStubByOrganisationsEinheitId(organisationsEinheitenId);
	}

	public String createVorgang(FormData formData, GrpcEingang eingang, VorgangServiceBlockingStub vorgangStub,
			BinaryFileServiceStub binaryFileStub) {
		return new VorgangCreator(formData, eingang, vorgangStub, binaryFileStub).create();
	}

	void finishStubConnections(List<ManagableStub<?>> stubs) {
		stubs.stream().filter(ManagableStub::isShutdownable).map(ClosableStub.class::cast).forEach(ClosableStub::close);
	}

	@RequiredArgsConstructor
	public class VorgangCreator {

		static final String VORGANG_ATTACHMENT_FIELD = "vorgangAttachment";
		static final String CALL_CONTEXT_CLIENT = "eingangAdpater";

		private final FormData formData;
		private final GrpcEingang eingang;
		private final VorgangServiceBlockingStub vorgangRemoteStub;
		private final BinaryFileServiceStub binaryFileRemoteStub;

		@Getter
		private String vorgangId;
		@Getter
		private List<IncomingFileGroup> uploadedAttachments;
		@Getter
		private List<IncomingFile> uploadedRepresentations;

		String create() {
			vorgangId = startCreation();

			uploadedAttachments = uploadAttachments();
			uploadedRepresentations = uploadRepresentations();

			finishCreation();
			return vorgangId;
		}

		String startCreation() {
			return vorgangRemoteStub.startCreation(buildStartCreationRequest(eingang)).getVorgangId();
		}

		private GrpcCreateVorgangRequest buildStartCreationRequest(GrpcEingang eingang) {
			var eingangWithoutFiles = eingang.toBuilder().clearAttachments().clearRepresentations().build();
			return GrpcCreateVorgangRequest.newBuilder().setEingang(eingangWithoutFiles).build();
		}

		List<IncomingFileGroup> uploadAttachments() {
			return formData.getAttachments().stream().map(this::uploadAttachment).toList();
		}

		private IncomingFileGroup uploadAttachment(IncomingFileGroup attachment) {
			var filesWithId = attachment.getFiles().stream().map(this::addIncomingFileId).toList();

			return IncomingFileGroup.builder().name(attachment.getName()).files(filesWithId).build();
		}

		List<IncomingFile> uploadRepresentations() {
			return formData.getRepresentations().stream().map(this::addIncomingFileId).toList();
		}

		private IncomingFile addIncomingFileId(IncomingFile file) {
			return file.toBuilder().id(uploadIncomingFile(file)).build();
		}

		String uploadIncomingFile(IncomingFile incomingFile) {
			var fileContentStream = incomingFile.getContentStream();

			var resultFuture = GrpcFileUploadUtils.createSender(this::buildChunkRequest, fileContentStream,
					this::buildCallStreamObserver)
					.withMetaData(buildMetaDataRequest(incomingFile))
					.send();

			return waitUntilFutureToComplete(resultFuture, fileContentStream).getFileId();
		}

		GrpcUploadBinaryFileRequest buildChunkRequest(byte[] bytes, Integer length) {
			return GrpcUploadBinaryFileRequest.newBuilder().setFileContent((ByteString.copyFrom(bytes, 0, length))).build();
		}

		private CallStreamObserver<GrpcUploadBinaryFileRequest> buildCallStreamObserver(
				StreamObserver<GrpcUploadBinaryFileResponse> responseObserver) {
			return (CallStreamObserver<GrpcUploadBinaryFileRequest>) binaryFileRemoteStub.uploadBinaryFileAsStream(responseObserver);
		}

		GrpcUploadBinaryFileRequest buildMetaDataRequest(IncomingFile ingomingFile) {
			return GrpcUploadBinaryFileRequest.newBuilder()
					.setMetadata(GrpcUploadBinaryFileMetaData.newBuilder()
							.setContext(GrpcCallContext.newBuilder().setClient(CALL_CONTEXT_CLIENT).build())
							.setVorgangId(getVorgangId())
							.setField(VORGANG_ATTACHMENT_FIELD)
							.setContentType(ingomingFile.getContentType())
							.setSize(ingomingFile.getSize())
							.setFileName(ingomingFile.getName()))
					.build();
		}

		GrpcUploadBinaryFileResponse waitUntilFutureToComplete(FileSender<GrpcUploadBinaryFileRequest, GrpcUploadBinaryFileResponse> fileSender,
				InputStream fileContentStream) {
			try {
				return fileSender.getResultFuture().get(2, TimeUnit.MINUTES);
			} catch (InterruptedException e) {
				Thread.currentThread().interrupt();
				fileSender.cancelOnError(e);
				throw new TechnicalException("Waiting for finishing upload was interrupted.", e);
			} catch (ExecutionException | TimeoutException e) {
				fileSender.cancelOnTimeout();
				throw new TechnicalException("Error / Timeout on uploading data.", e);
			} finally {
				IOUtils.closeQuietly(fileContentStream);
			}
		}

		String finishCreation() {
			return vorgangRemoteStub.finishCreation(buildFinishCreationRequest()).getMessage();
		}

		GrpcFinishCreationRequest buildFinishCreationRequest() {
			return GrpcFinishCreationRequest.newBuilder()
					.addAllAttachments(getAttachmentsWithoutContent())
					.addAllRepresentations(getRepresentationsWithoutContent())
					.setVorgangId(getVorgangId())
					.build();
		}

		private List<GrpcIncomingFileGroup> getAttachmentsWithoutContent() {
			return getUploadedAttachments().stream().map(this::toIncomingFileGroup).toList();
		}

		private GrpcIncomingFileGroup toIncomingFileGroup(IncomingFileGroup incomingFileGroup) {
			return GrpcIncomingFileGroup.newBuilder()
					.setName(incomingFileGroup.getName())
					.addAllFiles(incomingFileGroup.getFiles().stream().map(this::toIncomingFile).toList()).build();
		}

		private List<GrpcIncomingFile> getRepresentationsWithoutContent() {
			return getUploadedRepresentations().stream().map(this::toIncomingFile).toList();
		}

		private GrpcIncomingFile toIncomingFile(IncomingFile incomingFile) {
			return GrpcIncomingFile.newBuilder().clearContent()
					.setId(incomingFile.getId())
					.setContentType(incomingFile.getContentType())
					.setName(incomingFile.getName())
					.setSize(incomingFile.getSize()).build();
		}
	}
}