/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.eingang.router;

import java.io.File;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.List;
import java.util.Optional;
import java.util.stream.Stream;

import org.springframework.stereotype.Service;

import de.ozgcloud.eingang.common.formdata.FormData;
import de.ozgcloud.eingang.common.formdata.FormHeader;
import de.ozgcloud.eingang.common.formdata.IncomingFile;
import de.ozgcloud.eingang.common.formdata.IncomingFileGroup;
import de.ozgcloud.eingang.common.formdata.ServiceKonto;
import de.ozgcloud.eingang.common.formdata.ZustaendigeStelle;
import lombok.AllArgsConstructor;
import lombok.extern.log4j.Log4j2;

@AllArgsConstructor
@Service
@Log4j2
public class VorgangService {

	private final VorgangRemoteService remoteService;

	public String createVorgang(FormData formData) {
		var vorgangId = createVorgangForOrganisationsEinheitIds(getOrganisationsEinheitIds(formData), preserveConsistency(formData));
		cleanupFormDataFiles(formData);
		return vorgangId;
	}

	void cleanupFormDataFiles(FormData formData) {
		getFormDataFiles(formData)
				.map(IncomingFile::getFile)
				.map(File::toPath)
				.forEach(this::deleteIncomingFile);
	}

	void deleteIncomingFile(Path path) {
		try {
			Files.deleteIfExists(path);
		} catch (IOException e) {
			logErrorOnDeleteFailure(e);
		}
	}

	void logErrorOnDeleteFailure(Exception e) {
		LOG.error("Failed to delete temp-file of incoming file!", e);
	}

	Stream<IncomingFile> getFormDataFiles(FormData formData) {
		return Stream.concat(
				formData.getRepresentations().stream(),
				formData.getAttachments().stream()
						.map(IncomingFileGroup::getFiles)
						.flatMap(List::stream)
		);
	}

	String createVorgangForOrganisationsEinheitIds(List<String> organisationsEinheitIds, FormData preparedFormData) {
		return organisationsEinheitIds.isEmpty()
				? createVorgangOnMissingZustaendigeStelle(preparedFormData)
				: createMultipleVorgangs(preparedFormData, organisationsEinheitIds).getFirst();
	}

	String createVorgangOnMissingZustaendigeStelle(FormData formData) {
		return remoteService.createVorgang(formData, Optional.empty());
	}

	List<String> createMultipleVorgangs(FormData formData, List<String> organisationseinheitIds) {
		return organisationseinheitIds.stream()
				.map(oeId -> remoteService.createVorgang(formData, Optional.of(oeId)))
				.toList();
	}

	List<String> getOrganisationsEinheitIds(FormData formData) {
		return formData.getZustaendigeStelles().stream()
				.map(ZustaendigeStelle::getOrganisationseinheitenId)
				.toList();
	}

	FormData preserveConsistency(FormData formData) {
		if (hasNoPostfachAddress(formData)) {
			return clearServiceKonto(formData);
		}
		return formData;
	}

	private boolean hasNoPostfachAddress(FormData formData) {
		return Optional.ofNullable(formData.getHeader()).map(FormHeader::getServiceKonto).map(ServiceKonto::getPostfachAddresses).map(List::isEmpty)
				.orElse(true);
	}

	private FormData clearServiceKonto(FormData formData) {
		return formData.toBuilder().header(formData.getHeader().toBuilder().serviceKonto(null).build()).build();
	}
}