/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.eingang.common.zufi;

import static org.assertj.core.api.Assertions.*;
import static org.junit.jupiter.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;

import de.ozgcloud.eingang.router.errorhandling.MultipleOrganisationsEinheitenException;
import de.ozgcloud.eingang.router.errorhandling.UnknownOrganisationseinheitException;
import de.ozgcloud.zufi.grpc.organisationseinheit.GrpcGetByOrganisationsEinheitIdResponse;
import de.ozgcloud.zufi.grpc.organisationseinheit.OrganisationsEinheitServiceGrpc.OrganisationsEinheitServiceBlockingStub;

class OrganisationsEinheitRemoteServiceTest {

	@InjectMocks
	@Spy
	private OrganisationsEinheitRemoteService service;
	@Mock
	private OrganisationsEinheitServiceBlockingStub stub;
	@Mock
	private OrganisationsEinheitMapper mapper;

	@Nested
	class TestGetByOrganisationEinheitId {

		private final GrpcGetByOrganisationsEinheitIdResponse response = GrpcGetByOrganisationsEinheitIdResponseTestFactory.create();
		private final OrganisationsEinheit mappedOrganisationsEinheit = OrganisationsEinheitTestFactory.create();

		@BeforeEach
		void mock() {
			when(stub.getByOrganisationsEinheitId(any())).thenReturn(response);
			when(mapper.fromGrpc(any())).thenReturn(mappedOrganisationsEinheit);
			doNothing().when(service).verifyResponse(any(), any());
		}

		@Test
		void shouldCallRemoteService() {
			service.getByOrganisationEinheitId(OrganisationsEinheitTestFactory.ID);

			verify(stub).getByOrganisationsEinheitId(GrpcGetByOrganisationsEinheitIdRequestTestFactory.create());
		}

		@Test
		void shouldCallVerifyResponse() {
			service.getByOrganisationEinheitId(OrganisationsEinheitTestFactory.ID);

			verify(service).verifyResponse(OrganisationsEinheitTestFactory.ID, response);
		}

		@Test
		void shouldCallMapperAfterVerifyingResponse() {
			var order = inOrder(service, mapper);

			service.getByOrganisationEinheitId(OrganisationsEinheitTestFactory.ID);

			order.verify(service).verifyResponse(any(), any());
			order.verify(mapper).fromGrpc(GrpcGetByOrganisationsEinheitIdResponseTestFactory.ORGANISATIONS_EINHEIT);
		}

		@Test
		void shouldReturnMappedOrganisationsEinheit() {
			var organisationsEinheit = service.getByOrganisationEinheitId(OrganisationsEinheitTestFactory.ID);

			assertThat(organisationsEinheit).isEqualTo(mappedOrganisationsEinheit);
		}
	}

	@Nested
	class TestVerifyResponse {

		@Nested
		class TestOnMultipleOrganisationEinheitFound {

			@Test
			void shouldThrowMultipleOrganisationsEinheitenException() {
				var response = GrpcGetByOrganisationsEinheitIdResponseTestFactory.createBuilder()
						.addOrganisationsEinheiten(GrpcOrganisationsEinheitTestFactory.create())
						.build();

				assertThatExceptionOfType(MultipleOrganisationsEinheitenException.class)
						.isThrownBy(() -> service.verifyResponse(OrganisationsEinheitTestFactory.ID, response));

			}
		}

		@Nested
		class TestOnNoOrganisationEinheitFound {

			@Test
			void shouldThrowUnknownOrganisationseinheitException() {
				var response = GrpcGetByOrganisationsEinheitIdResponseTestFactory.createBuilder()
						.clearOrganisationsEinheiten()
						.build();

				assertThatExceptionOfType(UnknownOrganisationseinheitException.class)
						.isThrownBy(() -> service.verifyResponse(OrganisationsEinheitTestFactory.ID, response));

			}
		}

		@Nested
		class TestOnSingleOrganisationEinheitFound {

			@Test
			void shouldNotThrowException() {
				var response = GrpcGetByOrganisationsEinheitIdResponseTestFactory.create();

				assertDoesNotThrow(() -> service.verifyResponse(OrganisationsEinheitTestFactory.ID, response));
			}
		}
	}
}