/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.eingang.router;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.util.Map;

import org.apache.commons.lang3.StringUtils;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mapstruct.factory.Mappers;
import org.mockito.InjectMocks;
import org.mockito.Mock;

import de.ozgcloud.eingang.common.formdata.PostfachAddressTestFactory;
import de.ozgcloud.eingang.common.formdata.ServiceKontoTestFactory;
import de.ozgcloud.eingang.common.formdata.StringBasedIdentifier;
import de.ozgcloud.vorgang.common.GrpcObject;
import de.ozgcloud.vorgang.common.GrpcProperty;
import de.ozgcloud.vorgang.common.grpc.GrpcObjectMapper;
import de.ozgcloud.vorgang.vorgang.GrpcServiceKonto;

class ServiceKontoMapperTest {

	@InjectMocks
	private final ServiceKontoMapper mapper = Mappers.getMapper(ServiceKontoMapper.class);
	@Mock
	private GrpcObjectMapper grpcObjectMapper;

	@Nested
	class TestToGrpc {

		private final GrpcProperty grpcProperty = GrpcProperty.newBuilder()
				.setName(StringBasedIdentifier.POSTFACH_ID_FIELD)
				.addValue(PostfachAddressTestFactory.POSTFACH_ID)
				.build();
		private final GrpcObject grpcObject = GrpcObject.newBuilder()
				.addProperty(grpcProperty)
				.build();

		@BeforeEach
		void mockGrpcObjectMapper() {
			when(grpcObjectMapper.fromMap(any())).thenReturn(grpcObject);
		}

		@Test
		void shouldCallGrpcMapper() {
			getServiceKontoFromMappedEingang();

			verify(grpcObjectMapper).fromMap(any());
		}

		@Test
		void shouldBeExist() {
			var serviceKonto = getServiceKontoFromMappedEingang();

			assertThat(serviceKonto).isNotNull();
		}

		@Test
		void shouldContainsType() {
			var serviceKonto = getServiceKontoFromMappedEingang();

			assertThat(serviceKonto.getType()).isEqualTo(ServiceKontoTestFactory.TYPE);
		}

		@Test
		void shouldContainsDefaultTrustLevel() {
			var serviceKonto = mapper.toGrpc(ServiceKontoTestFactory.createBuilder().trustLevel(null).build());

			assertThat(serviceKonto.getTrustLevel()).isEqualTo(ServiceKontoMapper.DEFAULT_TRUST_LEVEL);
		}

		@Test
		void shouldContainsTrustLevel() {
			var serviceKonto = getServiceKontoFromMappedEingang();

			assertThat(serviceKonto.getTrustLevel()).isEqualTo(ServiceKontoTestFactory.TRUST_LEVEL);
		}

		@DisplayName("postfach address")
		@Nested
		class TestPostfachAddress {

			@Test
			void shouldHasSize() {
				var serviceKonto = getServiceKontoFromMappedEingang();

				assertThat(serviceKonto.getPostfachAddressesList()).hasSize(1);
			}

			@Test
			void shouldContainsVersion() {
				var serviceKonto = getServiceKontoFromMappedEingang();

				assertThat(serviceKonto.getPostfachAddressesList().get(0).getVersion()).isEqualTo(PostfachAddressTestFactory.VERSION);
			}

			@Test
			void shouldContainsIdentifier() {
				var serviceKonto = getServiceKontoFromMappedEingang();

				assertThat(serviceKonto.getPostfachAddressesList().get(0).getIdentifier().getPropertyList()).hasSize(1);

				var property = serviceKonto.getPostfachAddressesList().get(0).getIdentifier().getPropertyList().get(0);
				assertThat(property.getName()).isEqualTo(StringBasedIdentifier.POSTFACH_ID_FIELD);
				assertThat(property.getValue(0)).isEqualTo(PostfachAddressTestFactory.POSTFACH_ID);
			}

			@Test
			void shouldContainsType() {
				var serviceKonto = getServiceKontoFromMappedEingang();

				assertThat(serviceKonto.getPostfachAddressesList().get(0).getType()).isEqualTo(PostfachAddressTestFactory.POSTFACH_ADDRESS_TYPE);
			}
		}

		private GrpcServiceKonto getServiceKontoFromMappedEingang() {
			return mapper.toGrpc(ServiceKontoTestFactory.create());
		}
	}

	@Nested
	class TestFromGrpc {

		@BeforeEach
		void mockGrpcObjectMapper() {
			when(grpcObjectMapper.mapFromGrpc(any()))
					.thenReturn(Map.of(StringBasedIdentifier.POSTFACH_ID_FIELD, PostfachAddressTestFactory.POSTFACH_ID));
		}

		@Test
		void shouldCallGrpcMapper() {
			mapper.fromGrpc(GrpcServiceKontoTestFactory.create());

			verify(grpcObjectMapper).mapFromGrpc(GrpcPostfachAddressTestFactory.IDENTIFIER);
		}

		@Test
		void shouldMapToGrpc() {
			var serviceKonto = mapper.fromGrpc(GrpcServiceKontoTestFactory.create());

			assertThat(serviceKonto).usingRecursiveComparison().isEqualTo(ServiceKontoTestFactory.create());
		}

		@Test
		void shouldSetDefaultTrustLevel() {
			var serviceKonto = mapper.fromGrpc(GrpcServiceKontoTestFactory.createBuilder().setTrustLevel(StringUtils.EMPTY).build());

			assertThat(serviceKonto.getTrustLevel()).isEqualTo(ServiceKontoMapper.DEFAULT_TRUST_LEVEL);
		}
	}
}