/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.eingang.router;

import static de.ozgcloud.eingang.common.formdata.FormDataTestFactory.*;
import static java.util.Collections.*;
import static org.assertj.core.api.Assertions.*;
import static org.junit.jupiter.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.ArgumentMatchers.eq;
import static org.mockito.Mockito.*;

import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.List;
import java.util.Optional;
import java.util.stream.Stream;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;

import de.ozgcloud.eingang.common.formdata.FormData;
import de.ozgcloud.eingang.common.formdata.FormDataTestFactory;
import de.ozgcloud.eingang.common.formdata.FormHeader;
import de.ozgcloud.eingang.common.formdata.FormHeaderTestFactory;
import de.ozgcloud.eingang.common.formdata.IncomingFile;
import de.ozgcloud.eingang.common.formdata.IncomingFileTestFactory;
import de.ozgcloud.eingang.common.formdata.ServiceKonto;
import de.ozgcloud.eingang.common.formdata.ServiceKontoTestFactory;
import de.ozgcloud.eingang.common.formdata.ZustaendigeStelleTestFactory;

class VorgangServiceTest {

	@Spy
	@InjectMocks
	private VorgangService service;
	@Mock
	private VorgangRemoteService remoteService;

	private final FormData formData = FormDataTestFactory.create();

	@DisplayName("Create vorgang")
	@Nested
	class TestCreateVorgang {
		private final FormData preservedFormData = FormDataTestFactory.create();
		private final List<String> organisationseinheitIds = List.of(ZustaendigeStelleTestFactory.ORGANISATIONSEINHEIT_ID);

		@BeforeEach
		void mockEingangMapper() {
			doReturn(preservedFormData).when(service).preserveConsistency(any());
			doReturn(organisationseinheitIds).when(service).getOrganisationsEinheitIds(any());
			doReturn(VORGANG_ID).when(service).createVorgangForOrganisationsEinheitIds(any(), any());
			doNothing().when(service).cleanupFormDataFiles(any());
		}

		@Test
		void shouldPreserverConsistency() {
			callCreateVorgang();

			verify(service).preserveConsistency(formData);
		}

		@DisplayName("should call create vorgang for organisationseinheitIds")
		@Test
		void shouldCallCreateVorgangForOrganisationseinheitIds() {
			callCreateVorgang();

			verify(service).createVorgangForOrganisationsEinheitIds(organisationseinheitIds, preservedFormData);
		}

		@DisplayName("should call get organisationseinheit ids")
		@Test
		void shouldCallGetOrganisationseinheitIds() {
			callCreateVorgang();

			verify(service).getOrganisationsEinheitIds(formData);
		}

		@DisplayName("should return vorgang id")
		@Test
		void shouldReturn() {
			var vorgangId = callCreateVorgang();

			assertThat(vorgangId).isEqualTo(VORGANG_ID);
		}

		@DisplayName("should call cleanup form data files")
		@Test
		void shouldCallCleanupFormDataFiles() {
			callCreateVorgang();

			verify(service).cleanupFormDataFiles(formData);
		}

		private String callCreateVorgang() {
			return service.createVorgang(formData);
		}
	}

	@DisplayName("cleanup form data files")
	@Nested
	class TestCleanupFormDataFiles {
		private final IncomingFile incomingFile = IncomingFileTestFactory.create();

		@BeforeEach
		void mock() {
			doReturn(Stream.of(incomingFile)).when(service).getFormDataFiles(formData);
		}

		@DisplayName("should call delete incoming file")
		@Test
		void shouldCallDeleteIncomingFile() {
			service.cleanupFormDataFiles(formData);

			verify(service).deleteIncomingFile(incomingFile.getFile().toPath());
		}
	}

	@DisplayName("get form data files")
	@Nested
	class TestGetFormDataFiles {

		private final FormData formData = FormDataTestFactory.create();

		@DisplayName("should return attachments and representations")
		@Test
		void shouldReturnAttachmentsAndRepresentations() {
			var files = service.getFormDataFiles(formData).toList();

			assertThat(files).hasSize(FormDataTestFactory.ATTACHMENTS.size() + FormDataTestFactory.REPRESENTATIONS.size());
		}
	}

	@DisplayName("delete incoming file")
	@Nested
	class TestDeleteIncomingFile {

		@Mock
		private Path path;

		@DisplayName("should call deleteIfExists")
		@Test
		void shouldCallDeleteIfExists() {
			try (var staticMock = mockStatic(Files.class)) {
				service.deleteIncomingFile(path);

				staticMock.verify(() -> Files.deleteIfExists(path));
			}
		}

		@DisplayName("should return")
		@Test
		void shouldReturn() {
			try (var staticMock = mockStatic(Files.class)) {
				staticMock.when(() -> Files.deleteIfExists(path)).thenReturn(true);

				service.deleteIncomingFile(path);
			}
		}

		@DisplayName("should log on error")
		@Test
		void shouldLogOnError() {
			var exception = new IOException();
			try (var staticMock = mockStatic(Files.class)) {
				staticMock.when(() -> Files.deleteIfExists(path)).thenThrow(exception);

				service.deleteIncomingFile(path);

				verify(service).logErrorOnDeleteFailure(exception);
			}
		}

	}

	@DisplayName("create vorgang for organisationsEinheitIds")
	@Nested
	class TestCreateVorgangForOrganisationsEinheitIds {
		private final FormData preservedFormData = FormDataTestFactory.create();
		private final List<String> organisationseinheitIds = List.of(ZustaendigeStelleTestFactory.ORGANISATIONSEINHEIT_ID);

		@DisplayName("should call create multiple vorgangs")
		@Test
		void shouldCallCreateMultipleVorgangs() {
			service.createVorgangForOrganisationsEinheitIds(organisationseinheitIds, preservedFormData);

			verify(service).createMultipleVorgangs(preservedFormData, organisationseinheitIds);
		}

		@DisplayName("should return first vorgang id with multiple organisationseinheits")
		@Test
		void shouldReturnFirstVorgangIdWithMultipleOrganisationseinheits() {
			doReturn(List.of(VORGANG_ID)).when(service).createMultipleVorgangs(any(), any());

			var vorgangId = service.createVorgangForOrganisationsEinheitIds(organisationseinheitIds, preservedFormData);

			assertThat(vorgangId).isEqualTo(VORGANG_ID);
		}

		@DisplayName("should call create vorgang on missing zustaendige stelle if organisationseinheit is missing")
		@Test
		void shouldCallCreateVorgangOnMissingZustaendigeStelleIfOrganisationseinheitIsMissing() {
			service.createVorgangForOrganisationsEinheitIds(emptyList(), preservedFormData);

			verify(service).createVorgangOnMissingZustaendigeStelle(eq(preservedFormData));
		}

		@DisplayName("should return single vorgang id if organisationseinheit is missing")
		@Test
		void shouldReturnSingleVorgangIdIfOrganisationseinheitIsMissing() {
			doReturn(VORGANG_ID).when(service).createVorgangOnMissingZustaendigeStelle(any());

			var firstVorgangId = service.createVorgangForOrganisationsEinheitIds(emptyList(), preservedFormData);

			assertThat(firstVorgangId).isEqualTo(VORGANG_ID);
		}

	}

	@DisplayName("create vorgang on missing zustaendige stelle")
	@Nested
	class TestVorgangOnMissingZustaendigeStelle {
		@Mock
		private FormData formData;

		@BeforeEach
		void mock() {
			when(remoteService.createVorgang(any(), any())).thenReturn(VORGANG_ID);
		}

		@DisplayName("should call create vorgang")
		@Test
		void shouldCallCreateVorgang() {
			service.createVorgangOnMissingZustaendigeStelle(formData);

			verify(remoteService).createVorgang(formData, Optional.empty());
		}

		@DisplayName("should return")
		@Test
		void shouldReturn() {
			var vorgangId = service.createVorgangOnMissingZustaendigeStelle(formData);

			assertThat(vorgangId).isEqualTo(VORGANG_ID);
		}
	}

	@DisplayName("create multiple vorgangs")
	@Nested
	class TestCreateMultipleVorgangs {

		@Mock
		private FormData formData;

		@BeforeEach
		void mock() {
			when(remoteService.createVorgang(any(), any())).thenReturn(VORGANG_ID);
		}

		@DisplayName("should call create vorgang twice")
		@Test
		void shouldCallCreateVorgangTwice() {
			service.createMultipleVorgangs(formData, List.of(ZustaendigeStelleTestFactory.ORGANISATIONSEINHEIT_ID));

			verify(remoteService).createVorgang(formData, Optional.of(ZustaendigeStelleTestFactory.ORGANISATIONSEINHEIT_ID));
		}

		@DisplayName("should return")
		@Test
		void shouldReturn() {
			var vorgangIds = service.createMultipleVorgangs(formData, List.of(ZustaendigeStelleTestFactory.ORGANISATIONSEINHEIT_ID));

			assertThat(vorgangIds).containsExactly(VORGANG_ID);
		}
	}

	@DisplayName("get organisationseinheit ids")
	@Nested
	class TestGetOrganisationseinheitIds {
		private final FormData preservedFormData = FormDataTestFactory.create();

		@DisplayName("should return")
		@Test
		void shouldReturn() {
			var organisationseinheitIds = service.getOrganisationsEinheitIds(preservedFormData);

			assertThat(organisationseinheitIds).containsExactly(ZustaendigeStelleTestFactory.ORGANISATIONSEINHEIT_ID);
		}
	}

	@DisplayName("Preserve consistency")
	@Nested
	class TestPreserveConsistency {

		private final ServiceKonto serviceKonto = ServiceKontoTestFactory.createBuilder().clearPostfachAddresses().build();
		private final FormHeader header = FormHeaderTestFactory.createBuilder().serviceKonto(serviceKonto).build();
		private final FormData formData = FormDataTestFactory.createBuilder().header(header).build();

		@DisplayName("should set serviceKonto to null if no postfachAddress exists")
		@Test
		void shouldSetServiceKontoToNull() {
			var consistentFormData = service.preserveConsistency(formData);

			assertThat(consistentFormData.getHeader().getServiceKonto()).isNull();
		}

		@DisplayName("should keep serviceKonto if postfachAddress exists")
		@Test
		void shouldKeepServiceKonto() {
			var consistentFormData = service.preserveConsistency(FormDataTestFactory.create());

			assertThat(consistentFormData.getHeader().getServiceKonto()).isEqualTo(FormHeaderTestFactory.SERVICE_KONTO);
		}

		@DisplayName("should proceed without exception on missing serviceKonto")
		@Test
		void shouldProceedWithoutException() {
			var formDataWithoutServiceKonto = FormDataTestFactory.createBuilder()
					.header(FormHeaderTestFactory.createBuilder().serviceKonto(null).build()).build();

			assertDoesNotThrow(() -> service.preserveConsistency(formDataWithoutServiceKonto));
		}
	}
}