/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.eingang.semantik.enginebased.a12;

import java.io.File;
import java.io.IOException;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Optional;

import org.apache.commons.lang3.StringUtils;
import org.springframework.stereotype.Component;

import com.fasterxml.jackson.databind.ObjectMapper;

import de.ozgcloud.common.errorhandling.TechnicalException;
import de.ozgcloud.eingang.common.formdata.FormData;
import de.ozgcloud.eingang.common.formdata.FormData.FormDataControl;
import de.ozgcloud.eingang.common.formdata.FormData.Representations;
import de.ozgcloud.eingang.common.formdata.IncomingFile;
import lombok.RequiredArgsConstructor;
import lombok.extern.log4j.Log4j2;

@Log4j2
@Component
@RequiredArgsConstructor
public class A12FormDataMapper implements A12EngineBasedMapper {

	private final ObjectMapper objectMapper;

	@Override
	public FormData parseFormData(FormData formData) {
		return mergeReadFormData(formData, getFormDataFile(formData));
	}

	private FormData mergeReadFormData(FormData formData, Optional<IncomingFile> inFile) {
		return inFile.map(in -> readEingang(in.getFile()))
				.map(eingang -> eingang.getRoot())
				.map(formDataMap -> merge(formDataMap, formData.getFormData()))
				.map(formDataMap -> formData.toBuilder().formData(formDataMap).build())
				.orElse(formData);
	}

	Optional<IncomingFile> getFormDataFile(FormData formData) {
		var inFile = Optional.ofNullable(formData.getControl())
				.flatMap(FormDataControl::getRepresentations)
				.map(Representations::getPrimaryFormDataRepresentation)
				.flatMap(fileName -> getByName(fileName, formData.getRepresentations()));

		logWarnIfEmpty(inFile);
		return inFile;
	}

	A12Eingang readEingang(File formDataJson) {
		try {
			return objectMapper.readValue(formDataJson, A12Eingang.class);
		} catch (IOException e) {
			throw new TechnicalException("Error parsing JSON from A12 FormDataFile", e);
		}
	}

	void logWarnIfEmpty(Optional<IncomingFile> inFile) {
		if (inFile.isEmpty()) {
			LOG.warn("No primary Form Data Representation found.");
		}
	}

	Optional<IncomingFile> getByName(String name, List<IncomingFile> inFiles) {
		return inFiles.stream()
				.filter(inFile -> StringUtils.equals(name, inFile.getName()))
				.findFirst();
	}

	<A, B> Map<A, B> merge(Map<A, B> map1, Map<A, B> map2) {
		var map = new HashMap<>(map1);
		map.putAll(map2);
		return Collections.unmodifiableMap(map);
	}
}
