/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.eingang.semantik.common;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.util.List;
import java.util.Map;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;
import org.mockito.Spy;

import de.ozgcloud.eingang.common.formdata.FormData;
import de.ozgcloud.eingang.common.formdata.FormDataUtils;
import de.ozgcloud.eingang.common.formdata.PostfachAddressTestFactory;
import de.ozgcloud.eingang.common.formdata.ServiceKonto;
import de.ozgcloud.eingang.common.formdata.ServiceKonto.TrustLevel;
import de.ozgcloud.eingang.common.formdata.ServiceKontoTestFactory;
import de.ozgcloud.eingang.common.formdata.StringBasedIdentifier;
import de.ozgcloud.eingang.common.formdata.ServiceKonto.PostfachAddress;

class ServiceKontoFactoryTest {

	@Spy
	@InjectMocks
	private ServiceKontoFactory factory;

	@DisplayName("OSI service konto")
	@Nested
	class TestOsiServiceKonto {

		private static final FormData FORM_DATA = FormDataTestFactory.create();

		@DisplayName("with configured postfach")
		@Nested
		class TestWithConfiguredPostfach {

			private static final PostfachAddress POSTFACH_ADDRESS = PostfachAddressTestFactory.create();

			@BeforeEach
			void mockBuildPostfachAddresses() {
				doReturn(List.of(POSTFACH_ADDRESS)).when(factory).buildPostfachAddresses(any(), any());
			}

			@Test
			void shouldContainsType() {
				var serviceKonto = getServiceKonto(FORM_DATA);

				assertThat(serviceKonto.getType()).isEqualTo(ServiceKontoFactory.POSTFACH_TYPE_OSI);
			}

			@Test
			void shouldContainsPostfachAddresses() {
				var serviceKonto = getServiceKonto(FORM_DATA);

				assertThat(serviceKonto.getPostfachAddresses()).hasSize(1);
				assertThat(serviceKonto.getPostfachAddresses().get(0)).isEqualTo(POSTFACH_ADDRESS);
			}

			@Test
			void shouldBuildPostfachAddresses() {
				getServiceKonto(FORM_DATA);

				verify(factory).buildPostfachAddresses(any(), any());
			}
		}

		private ServiceKonto getServiceKonto(FormData formData) {
			return factory.buildOsiServiceKonto(FormDataTestFactory.POSTFACH_NAME_ID, formData);
		}

		@DisplayName("postfach addresses")
		@Nested
		class TestBuildPostfachAddresses {

			@DisplayName("with rest_response_name")
			@Nested
			class TestWithRestResponseName {

				@Test
				void shouldCallBuildAddresses() {
					getPostfachAddresses();

					verify(factory).buildOsiPostfachV1Address(any(), anyInt());
				}

				@Test
				void shouldReturnPostfachAddresses() {
					var addresses = getPostfachAddresses();

					assertThat(addresses).hasSize(1);
					assertThat(addresses.get(0).getIdentifier()).isInstanceOf(StringBasedIdentifier.class);
					assertThat(((StringBasedIdentifier) addresses.get(0).getIdentifier()).getPostfachId())
							.isEqualTo(FormDataTestFactory.POSTFACH_NAME_ID);
					assertThat(addresses.get(0).getVersion()).isEqualTo(ServiceKontoFactory.POSTFACH_VERSION);
					assertThat(addresses.get(0).getType()).isEqualTo(PostfachAddressTestFactory.POSTFACH_ADDRESS_TYPE);
				}

				private List<PostfachAddress> getPostfachAddresses() {
					return buildServiceKonto(FORM_DATA).getPostfachAddresses();
				}
			}

			@DisplayName("without rest_response_name")
			@Nested
			class TestWithoutRestResponseName {

				private static final FormData FORM_DATA_WITHOUT_REST_RESPONSE_NAME = FormDataUtils.from(FORM_DATA)
						.remove(ServiceKontoFactory.REST_RESPONSE_NAME).build();

				@Test
				void shouldBuildDefault() {
					getPostfachAddresses();

					verify(factory).buildDefault(FormDataTestFactory.POSTFACH_NAME_ID);
				}

				@Test
				void shouldReturnPostfachAddresses() {
					var addresses = getPostfachAddresses();

					assertThat(addresses).hasSize(1);

					assertThat(addresses.get(0).getIdentifier()).isInstanceOf(StringBasedIdentifier.class);

					assertThat(((StringBasedIdentifier) addresses.get(0).getIdentifier()).getPostfachId())
							.isEqualTo(FormDataTestFactory.POSTFACH_NAME_ID);
					assertThat(addresses.get(0).getVersion()).isEqualTo(ServiceKontoFactory.POSTFACH_VERSION);
					assertThat(addresses.get(0).getType()).isEqualTo(1);
				}

				private List<PostfachAddress> getPostfachAddresses() {
					return buildServiceKonto(FORM_DATA_WITHOUT_REST_RESPONSE_NAME).getPostfachAddresses();
				}
			}

			private ServiceKonto buildServiceKonto(FormData formData) {
				return factory.buildOsiServiceKonto(FormDataTestFactory.POSTFACH_NAME_ID, formData);
			}
		}
	}

	@Nested
	class TestCreateBayernIdServicekonto {

		private final Map<String, Object> formDataHeaders = Map.of(
				ServiceKontoFactory.KEY_BAYERN_ID_POSTFACH_ID, PostfachAddressTestFactory.POSTFACH_ID,
				ServiceKontoFactory.KEY_BAYERN_ID_TRUST_LEVEL, ServiceKontoTestFactory.TRUST_LEVEL
		);

		@DisplayName("should return empty when headers map is null")
		@Test
		void shouldReturnEmptyWhenNull() {
			var serviceKonto = factory.createBayernIdServiceKonto(null);

			assertThat(serviceKonto).isEmpty();
		}

		@DisplayName("should return empty when postfach id is missing")
		@Test
		void shouldReturnEmptyWhenPostfachIdIsMissing() {
			var serviceKonto = factory.createBayernIdServiceKonto(Map.of());

			assertThat(serviceKonto).isEmpty();
		}

		@Test
		void shouldCallIsValidTrustLevel() {
			factory.createBayernIdServiceKonto(formDataHeaders);

			verify(factory).isValidTrustLevel(formDataHeaders);
		}

		@Test
		void shouldCallBuildBayernIdServiceKonto() {
			factory.createBayernIdServiceKonto(formDataHeaders);

			verify(factory).buildBayernIdServiceKonto(formDataHeaders);
		}

		@Test
		void shouldReturnServiceKonto() {
			var serviceKonto = ServiceKonto.builder().build();
			doReturn(serviceKonto).when(factory).buildBayernIdServiceKonto(any());

			var result = factory.createBayernIdServiceKonto(formDataHeaders);

			assertThat(result).contains(serviceKonto);
		}

		@DisplayName("should return empty when trust level has unexpected value")
		@Test
		void shouldReturnEmptyWhenTrustLevelCorrupted() {
			doReturn(false).when(factory).isValidTrustLevel(any());

			var serviceKonto = factory.createBayernIdServiceKonto(formDataHeaders);

			assertThat(serviceKonto).isEmpty();
		}
	}

	@Nested
	class TestIsValidTrustLevel {

		@Test
		void shouldCallExists() {
			try (var trustLevelMock = mockStatic(TrustLevel.class)) {
				isValidTrustLevel();

				trustLevelMock.verify(() -> TrustLevel.exists(ServiceKontoTestFactory.TRUST_LEVEL));
			}
		}

		@Test
		void shouldReturnValue() {
			try (var trustLevelMock = mockStatic(TrustLevel.class)) {
				trustLevelMock.when(() -> TrustLevel.exists(any())).thenReturn(true);

				var result = isValidTrustLevel();

				assertThat(result).isTrue();
			}
		}

		private boolean isValidTrustLevel() {
			return factory.isValidTrustLevel(Map.of(ServiceKontoFactory.KEY_BAYERN_ID_TRUST_LEVEL, ServiceKontoTestFactory.TRUST_LEVEL));
		}
	}

	@Nested
	class TestBuildBayernIdServiceKonto {

		private static final String POSTFACH_ID = "postfach-id";
		private static final PostfachAddress POSTFACH_ADDRESS = PostfachAddressTestFactory.create();
		private static final String TRUST_LEVEL = "STORK-QAA-Level-2";

		private final Map<String, Object> formDataHeaders = Map.of(
				ServiceKontoFactory.KEY_BAYERN_ID_POSTFACH_ID, POSTFACH_ID,
				ServiceKontoFactory.KEY_BAYERN_ID_TRUST_LEVEL, TRUST_LEVEL);

		@Test
		void shouldSetType() {
			var serviceKonto = buildBayernIdServiceKonto();

			assertThat(serviceKonto.getType()).isEqualTo(ServiceKontoFactory.POSTFACH_TYPE_BAYERN_ID);
		}

		@Test
		void shouldCallBuildPostfachAddress() {
			buildBayernIdServiceKonto();

			verify(factory).buildPostfachAddress(POSTFACH_ID);
		}

		@Test
		void shouldSetPostfachAddress() {
			doReturn(POSTFACH_ADDRESS).when(factory).buildPostfachAddress(any());

			var serviceKonto = buildBayernIdServiceKonto();

			assertThat(serviceKonto.getPostfachAddresses()).containsOnly(POSTFACH_ADDRESS);
		}

		@Test
		void shouldCallGetTrustLevel() {
			buildBayernIdServiceKonto();

			verify(factory).getTrustLevel(formDataHeaders);
		}

		@Test
		void shouldSetTrustLevel() {
			var serviceKonto = buildBayernIdServiceKonto();

			assertThat(serviceKonto.getTrustLevel()).isEqualTo(TRUST_LEVEL);
		}

		private ServiceKonto buildBayernIdServiceKonto() {
			return factory.buildBayernIdServiceKonto(formDataHeaders);
		}
	}

	@Nested
	class TestGetTrustLevel {

		private final Map<String, Object> formDataHeaders = Map.of(ServiceKontoFactory.KEY_BAYERN_ID_TRUST_LEVEL,
				ServiceKontoTestFactory.TRUST_LEVEL);

		@Test
		void shouldReturnTrustLevel() {
			var trustLevel = factory.getTrustLevel(formDataHeaders);

			assertThat(trustLevel).isEqualTo(ServiceKontoTestFactory.TRUST_LEVEL);
		}

	}

}