/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.eingang.semantik.common;

import static org.assertj.core.api.Assertions.*;
import static org.junit.jupiter.api.Assertions.*;
import static org.mockito.Mockito.*;

import java.io.ByteArrayInputStream;
import java.io.File;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.List;
import java.util.function.Predicate;
import java.util.zip.ZipException;
import java.util.zip.ZipInputStream;

import org.apache.commons.lang3.StringUtils;
import org.apache.commons.lang3.tuple.Pair;
import org.junit.jupiter.api.Assertions;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.Spy;
import org.springframework.util.MimeTypeUtils;

import de.ozgcloud.common.test.TestUtils;
import de.ozgcloud.eingang.common.formdata.IncomingFile;
import lombok.SneakyThrows;

class ZipAttachmentReaderTest {
	private static final String TMP_DIRECTORY_PATH = System.getProperty("java.io.tmpdir");
	private static final String ZIP_1_FILE_NAME = "attachment-1file.zip";
	private static final String ZIP_ENCRYPTED = "attachment-encrypted.zip";

	@Spy
	private ZipAttachmentReader reader;

	@Nested
	class TestReadZipAttachment {

		@Test
		@SneakyThrows
		@DisplayName("should save original zip to file system")
		void shouldSaveOriginalZip() {
			cleanupTempFiles();

			createZipAttachment(ZIP_1_FILE_NAME);

			verifySourceFileSavedInTmpDirectory();
		}

		@SneakyThrows
		private void verifySourceFileSavedInTmpDirectory() {
			List<Path> foundFiles = Files.find(Path.of(getTmpDirectoryPath()), 1,
					((path, basicFileAttributes) -> path.getFileName().toString().startsWith(ZipAttachmentReader.SOURCE_ZIP_PREFIX)))
					.toList();
			assertThat(foundFiles).hasSize(1).first();
		}

		@Test
		@DisplayName("should return readable input stream for source zip file")
		void shouldReturnSourceStream() {
			var expectedContent = TestUtils.loadFile(ZIP_1_FILE_NAME);

			var sourceZipAsStream = createZipAttachment(ZIP_1_FILE_NAME).getSourceZipAsStream();

			assertThat(sourceZipAsStream).hasSameContentAs(expectedContent);
		}

		@Test
		@DisplayName("should throw exception when reading encrypted ZIP")
		void shouldFailSilentByEncryptedZip() {
			var zipAttachment = createZipAttachment(ZIP_ENCRYPTED);

			assertThrows(ReadZipException.class, zipAttachment::readContent);
		}

		private static ZipAttachmentReader createZipAttachment(String fileName) {
			return ZipAttachmentReader.from(TestUtils.loadFile(fileName), fileName);
		}
	}

	@Nested
	class TestReadZip {

		private static final String ZIP_2_FILE_NAME = "attachment-2files.zip";
		private static final String content_file_0_name = "zip-file-0.txt";
		private static final long content_file_0_size = getFileSize("zip-file-0.txt");
		private static final String content_file_1_name = "zip-file-1.txt";
		private static final long content_file_1_size = getFileSize("zip-file-1.txt");

		private static Pair<String, Long> createContentPair(String fileName) {
			var size = switch (fileName) {
				case content_file_0_name -> content_file_0_size;
				case content_file_1_name -> content_file_1_size;
				default -> getFileSize(fileName);
			};
			return Pair.of(fileName, size);
		}

		@Test
		@SneakyThrows
		void shouldReadAllZipEntries() {
			var zipContent = new ZipAttachmentReader().readContent(loadZip(ZIP_2_FILE_NAME)).stream()
					.map(e -> Pair.of(e.getName(), e.getSize())).toList();

			assertThat(zipContent).containsExactlyInAnyOrder(createContentPair(content_file_0_name), createContentPair(content_file_1_name));
		}

		@Test
		@SneakyThrows
		void shouldReadZipContent() {
			var attachmentContentList = new ZipAttachmentReader().readContent(loadZip(ZIP_1_FILE_NAME));

			assertThat(attachmentContentList).hasSize(1);
			var contentEntry = attachmentContentList.get(0);
			assertThat(contentEntry.getName()).isEqualTo(content_file_0_name);
			assertThat(contentEntry.getSize()).isEqualTo(content_file_0_size);
			assertThat(contentEntry.getContentStream()).hasSameContentAs(TestUtils.loadFile(content_file_0_name));
		}

		@Test
		@SneakyThrows
		void shouldSkipFolders() {
			cleanupTempFiles();

			var zipContent = new ZipAttachmentReader().readContent(loadZip("attachment-empty.zip"));

			assertThat(zipContent).isEmpty();
			assertTrue(noFilesWithSuffixInTempDirectory());
		}

		@Test
		@DisplayName("should delete all temporary files after last reading of inputstream")
		@SneakyThrows
		void shouldDeleteContentFilesOnFinalRead() {
			cleanupTempFiles();
			var contentEntries = new ZipAttachmentReader().readContent(loadZip(ZIP_2_FILE_NAME));

			contentEntries.forEach(this::closeInputStreamFinalRead);

			assertTrue(noFilesWithSuffixInTempDirectory());
		}

		@SneakyThrows
		private void closeInputStreamFinalRead(IncomingFile entry) {
			entry.getContentStreamForFinalRead().close();
		}

		@Test
		@DisplayName("should return readable input stream for source zip if cannot extract content")
		void shouldReturnSourceStreamByError() {
			var attachmentContent = new byte[] { 0, 1, 2, 3 };
			var attachment = ZipAttachmentReader.from(new ByteArrayInputStream(attachmentContent), "invalid");

			assertThrows(ReadZipException.class, attachment::readContent);

			assertThat(attachment.getSourceZipAsStream()).hasSameContentAs(new ByteArrayInputStream(attachmentContent));
		}

		@Test
		@DisplayName("should throw exception if ZIP is invalid.")
		void shouldFailSilentByError() {
			var invalidZip = new ByteArrayInputStream(new byte[] { 0, 1, 2, 3 });

			var zipAttachment = ZipAttachmentReader.from(invalidZip, "invalid");

			assertThrows(ReadZipException.class, zipAttachment::readContent);
		}

		@Test
		@SneakyThrows
		void shouldFailByEncryptedZip() {
			Assertions.assertThrows(ZipException.class, () -> new ZipAttachmentReader().readContent(loadZip(ZIP_ENCRYPTED)));
		}

	}

	@SneakyThrows
	private static ZipInputStream loadZip(String name) {
		return new ZipInputStream(TestUtils.loadFile(name));
	}

	@Nested
	class TestZipBombs {

		private static final String ZIP_BOMB_WITH_BIG_NULL_FILE_CONTENT = "zipbombs/filewithnulls.dat.zip";
		private static final String ZIP_BOMB_WITH_MANY_FILES = "zipbombs/filewithmanyfiles.dat.zip";

		@Test
		void shouldFailOnExtremCompressionRatio() {
			var zip = loadZip(ZIP_BOMB_WITH_BIG_NULL_FILE_CONTENT);

			ReadZipException exception = assertThrows(ReadZipException.class, () -> reader.readContent(zip));

			assertThat(exception.getMessage()).contains("Ratio between compressed and uncompressed data is highly suspicious");
		}

		@Test
		@SneakyThrows
		void shouldFailOnTotalExtractedSize() {
			var zip = loadZip(ZIP_1_FILE_NAME);
			reader.readContent(zip);

			verify(reader).checkTotalExtractedSize(157);
		}

		@Test
		void shouldFailOnTotalZipEntries() {
			var zip = loadZip(ZIP_BOMB_WITH_MANY_FILES);

			ReadZipException exception = assertThrows(ReadZipException.class, () -> reader.readContent(zip));

			assertThat(exception.getMessage()).contains("Total entries in zip file exceeded");
		}
	}

	@Nested
	class TestSaveFiles {

		@Test
		@SneakyThrows
		@DisplayName("should save file in temporary folder")
		void shouldSaveFile() {
			var systemTmpPathWithoutLastSlash = getTmpDirectoryPath();

			var resultFile = reader.createLocalTempFile();

			assertThat(resultFile).hasParent(systemTmpPathWithoutLastSlash);
		}

		@Test
		@SneakyThrows
		@DisplayName("should save file with specific prefix and suffix")
		void shouldSaveFileByName() {
			var resultFileName = new ZipAttachmentReader()
					.createLocalTempFile().getName();

			assertThat(resultFileName)
					.startsWith(ZipAttachmentReader.TARGET_ATTACHMENT_PREFIX)
					.endsWith(ZipAttachmentReader.TMP_FILE_SUFFIX);
		}
	}

	@Nested
	class TestContentType {

		@Test
		void shouldReturnDefaultWhenNullString() {
			assertThrows(NullPointerException.class, () -> reader.getContentType(null));
		}

		@Test
		void shouldReturnDefaultWhenEmptyString() {
			var contentType = new ZipAttachmentReader().getContentType(StringUtils.EMPTY);

			assertThat(contentType).isEqualTo(MimeTypeUtils.APPLICATION_OCTET_STREAM_VALUE);
		}

		@Test
		void shouldReturnDefaultWhenSpaceString() {
			var contentType = new ZipAttachmentReader().getContentType(StringUtils.SPACE);

			assertThat(contentType).isEqualTo(MimeTypeUtils.APPLICATION_OCTET_STREAM_VALUE);
		}

		@Test
		void shouldGetContentType() {
			var fileNames = List.of("1.xml", "2.txt");

			var contentTypes = fileNames.stream().map(new ZipAttachmentReader()::getContentType).toList();

			assertThat(contentTypes).containsExactlyInAnyOrder(MimeTypeUtils.APPLICATION_XML_VALUE, MimeTypeUtils.TEXT_PLAIN_VALUE);
		}
	}

	private static final Predicate<Path> hasNameSuffix = p -> p.getFileName().toString().endsWith(ZipAttachmentReader.TMP_FILE_SUFFIX);

	@SneakyThrows
	private static void cleanupTempFiles() {
		Files.walk(Path.of(TMP_DIRECTORY_PATH), 1).filter(hasNameSuffix).map(Path::toFile).forEach(File::delete);
	}

	@SneakyThrows
	private static boolean noFilesWithSuffixInTempDirectory() {
		return Files.walk(Path.of(TMP_DIRECTORY_PATH), 1).noneMatch(hasNameSuffix);
	}

	@SneakyThrows
	private static long getFileSize(String fileName) {
		Path filePath = Path.of(ZipAttachmentReaderTest.class.getClassLoader().getResource(fileName).toURI());
		return Files.size(filePath);
	}

	private static String getTmpDirectoryPath() {
		return TMP_DIRECTORY_PATH.endsWith("/") ? TMP_DIRECTORY_PATH.substring(0, TMP_DIRECTORY_PATH.length() - 1) : TMP_DIRECTORY_PATH;
	}
}
