/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.eingang.semantik.enginebased;

import static de.ozgcloud.eingang.common.formdata.FormDataTestFactory.*;
import static org.assertj.core.api.Assertions.*;

import java.util.List;
import java.util.Map;

import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;

import de.ozgcloud.eingang.common.formdata.FormData;
import de.ozgcloud.eingang.common.formdata.FormDataTestFactory;
import de.ozgcloud.eingang.common.formdata.IncomingFileGroup;
import de.ozgcloud.eingang.common.formdata.IncomingFileGroupTestFactory;
import de.ozgcloud.eingang.common.formdata.IncomingFileTestFactory;

class FilesMapperHelperTest {

	@Nested
	class TestExtractData {

		@Test
		void shouldReturnMappedFiles() {
			Map<String, String> expectedResult = Map.of(SIMPLE_VALUE_KEY, SIMPLE_VALUE);
			var mappedFilesMap = Map.of(FilesMapperHelper.FIELD_NAME_MAPPED_FILES, (Object) expectedResult);

			var extractedMappedFiles = FilesMapperHelper.getMappedFiles(FormDataTestFactory.withFormDataMaps(mappedFilesMap));

			assertThat(extractedMappedFiles.get()).isEqualTo(expectedResult);
		}

		@Test
		void shouldHandleMissingFiles() {
			var extractedMappedFiles = FilesMapperHelper.getMappedFiles(FormDataTestFactory.withFormDataMaps(Map.of(SIMPLE_VALUE_KEY, SIMPLE_VALUE)));

			assertThat(extractedMappedFiles).isEmpty();
		}

		@Test
		void shouldReturnAttachedFileGroups() {
			var incomingFileGroups = List.of(IncomingFileGroupTestFactory.create());
			Map<String, Object> mappedFilesMap = Map.of(FilesMapperHelper.FIELD_NAME_MAPPED_FILES,
					Map.of(FilesMapperHelper.ATTACHMENTS, incomingFileGroups));

			var extractedAttachments = FilesMapperHelper.getAttachedFileGroups(FormDataTestFactory.withFormDataMaps(mappedFilesMap));

			assertThat(extractedAttachments.get()).isEqualTo(incomingFileGroups);
		}

		@Nested
		class getAttachedFiledGroups {
			@Test
			void shouldHandleMissingFileGroups() {
				Map<String, Object> mappedFilesMap = Map.of(FilesMapperHelper.FIELD_NAME_MAPPED_FILES, Map.of(SIMPLE_VALUE_KEY, SIMPLE_VALUE));

				var extractedFileGroups = FilesMapperHelper.getAttachedFileGroups(FormDataTestFactory.withFormDataMaps(mappedFilesMap));

				assertThat(extractedFileGroups).isEmpty();
			}
		}

		@Test
		void souldReturnRepresentations() {
			var incomingFileGroups = List.of(IncomingFileTestFactory.create());
			Map<String, Object> mappedFilesMap = Map.of(FilesMapperHelper.FIELD_NAME_MAPPED_FILES,
					Map.of(FilesMapperHelper.REPRESENTATIONS, incomingFileGroups));

			var extractedRepresentations = FilesMapperHelper.getRepresentations(FormDataTestFactory.withFormDataMaps(mappedFilesMap));

			assertThat(extractedRepresentations.get()).isEqualTo(incomingFileGroups);
		}

		@Test
		void shouldHandleMissingFileRepresentations() {
			Map<String, Object> mappedFilesMap = Map.of(FilesMapperHelper.FIELD_NAME_MAPPED_FILES, Map.of(SIMPLE_VALUE_KEY, SIMPLE_VALUE));

			var extractedRepresentations = FilesMapperHelper.getRepresentations(FormDataTestFactory.withFormDataMaps(mappedFilesMap));

			assertThat(extractedRepresentations).isEmpty();
		}
	}

	@Nested
	class TestAttachmentCount {

		@Test
		void shouldCountEmptyList() {
			var counter = FilesMapperHelper.countAttachedFiles(List.of());

			assertThat(counter).isZero();
		}

		@Test
		void shouldCountNoAttachedFiles() {
			var counter = FilesMapperHelper.countAttachedFiles(List.of(IncomingFileGroupTestFactory.createBuilder().clearFiles().build()));

			assertThat(counter).isZero();
		}

		@Test
		void shouldCountOneAttachment() {
			var fileGroup = List.of(IncomingFileGroupTestFactory.createBuilder().file(IncomingFileTestFactory.create()).build());

			var counter = FilesMapperHelper.countAttachedFiles(fileGroup);

			assertThat(counter).isEqualTo(2);
		}

		@Test
		void shouldCountAllAttachments() {
			var counter = FilesMapperHelper.countAttachedFiles(createFileGroupsWith4Files());

			assertThat(counter).isEqualTo(4);
		}

		private static List<IncomingFileGroup> createFileGroupsWith4Files() {
			return List.of(IncomingFileGroupTestFactory.createBuilder()
					.clearFiles()
					.files(List.of(IncomingFileTestFactory.create(), IncomingFileTestFactory.create(), IncomingFileTestFactory.create())).build(),
					IncomingFileGroupTestFactory.create());

		}
	}

	@Nested
	class TestRemoveProcessedData {

		@Test
		@DisplayName("should remove processed mapped files from raw form data")
		void shouldRemoveProcessedDataOnly() {
			var rawFormData = Map.of(SIMPLE_VALUE_KEY, SIMPLE_VALUE, FilesMapperHelper.FIELD_NAME_MAPPED_FILES, new Object());

			var cleanedFormData = FilesMapperHelper.removeProcessedData(withFormDataMaps(rawFormData)).getFormData();

			assertThat(cleanedFormData)
					.doesNotContainKey(FilesMapperHelper.FIELD_NAME_MAPPED_FILES)
					.containsEntry(SIMPLE_VALUE_KEY, SIMPLE_VALUE);
		}

		@Test
		@DisplayName("should not change any other data")
		void shouldNotChangeOtherFields() {
			FormData formData = create();

			var cleanedFormData = FilesMapperHelper.removeProcessedData(formData);

			assertThat(cleanedFormData.getId()).isEqualTo(formData.getId());
			assertThat(cleanedFormData.getHeader()).isEqualTo(formData.getHeader());
			assertThat(cleanedFormData.getZustaendigeStelles()).isEqualTo(formData.getZustaendigeStelles());
			assertThat(cleanedFormData.getAntragsteller()).isEqualTo(formData.getAntragsteller());
			assertThat(cleanedFormData.getNumberOfAttachments()).isEqualTo(formData.getNumberOfAttachments());
			assertThat(cleanedFormData.getAttachments()).isEqualTo(formData.getAttachments());
			assertThat(cleanedFormData.getNumberOfRepresentations()).isEqualTo(formData.getNumberOfRepresentations());
			assertThat(cleanedFormData.getRepresentations()).isEqualTo(formData.getRepresentations());
			assertThat(cleanedFormData.getFormData()).containsAllEntriesOf(formData.getFormData());
		}
	}
}