/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.eingang.semantik.enginebased.formcycle;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.Mockito.*;

import java.util.HashMap;
import java.util.Map;
import java.util.Optional;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;

import de.ozgcloud.eingang.common.formdata.Antragsteller;
import de.ozgcloud.eingang.common.formdata.AntragstellerTestFactory;
import de.ozgcloud.eingang.common.formdata.FormData;
import de.ozgcloud.eingang.common.formdata.FormDataTestFactory;

class FormcycleAntragstellerMapperTest {

	@Spy
	@InjectMocks
	private FormcycleAntragstellerMapper mapper;

	private static final String VALUE_KEY = "value";
	private static final String LABEL_KEY = "label";

	private Map<String, Object> antragstellerData = new HashMap<>();
	private Map<String, Object> antragstellerMap = Map.of("fsBKAllDaten", Map.of(VALUE_KEY, antragstellerData));

	@Nested
	class TestParseFormData {

		@Mock
		private Antragsteller antragsteller;

		@Test
		void shouldCallGetAntragsteller() {
			mapper.parseFormData(buildFormData());

			verify(mapper).getAntragstellerData(antragstellerMap);
		}

		@Test
		void shouldCallBuildAntragsteller() {
			doReturn(Optional.of(antragstellerData)).when(mapper).getAntragstellerData(anyMap());

			mapper.parseFormData(buildFormData());

			verify(mapper).buildAntragsteller(antragstellerData);
		}

		@Test
		void shouldSetAntragsteller() {
			doReturn(antragsteller).when(mapper).buildAntragsteller(anyMap());

			var formData = mapper.parseFormData(buildFormData());

			assertThat(formData.getAntragsteller()).isSameAs(antragsteller);
		}

		@Test
		void shouldReturnSameFormData() {
			var formData = FormDataTestFactory.create();

			var result = mapper.parseFormData(formData);

			assertThat(result).isSameAs(formData);
		}

		FormData buildFormData() {
			return FormData.builder().formData(antragstellerMap).build();
		}
	}

	@Nested
	class TestGetAntragstellerData {

		@Test
		void shouldReturnEmpty() {
			Map<String, Object> formDataMap = Map.of("key", "value");

			var result = mapper.getAntragstellerData(formDataMap);

			assertThat(result).isEmpty();
		}

		@Test
		void shouldReturnAntragstellerData() {
			var result = mapper.getAntragstellerData(antragstellerMap);

			assertThat(result).contains(antragstellerData);
		}
	}

	@Nested
	class TestBuildAntragsteller {

		@Test
		void shouldSetAnrede() {
			antragstellerData.put("tfAntragstellerAnrede", Map.of(VALUE_KEY, AntragstellerTestFactory.ANREDE, LABEL_KEY, "Anrede"));

			var result = buildAntragsteller();

			assertThat(result.getAnrede()).isEqualTo(AntragstellerTestFactory.ANREDE);
		}

		@Test
		void shouldSetVorname() {
			antragstellerData.put("tfAntragstellerVorname", Map.of(LABEL_KEY, "Vorname", VALUE_KEY, AntragstellerTestFactory.VORNAME));

			var result = buildAntragsteller();

			assertThat(result.getVorname()).isEqualTo(AntragstellerTestFactory.VORNAME);
		}

		@Test
		void shouldSetNachname() {
			antragstellerData.put("tfAntragstellerName", Map.of(VALUE_KEY, AntragstellerTestFactory.NACHNAME, LABEL_KEY, "Nachname"));

			var result = buildAntragsteller();

			assertThat(result.getNachname()).isEqualTo(AntragstellerTestFactory.NACHNAME);
		}

		@Test
		void shouldSetGeburtsname() {
			antragstellerData.put("tfAntragstellerGeburtsname", Map.of(VALUE_KEY, AntragstellerTestFactory.GEBURTSNAME, LABEL_KEY, "Geburtsname"));

			var result = buildAntragsteller();

			assertThat(result.getGeburtsname()).isEqualTo(AntragstellerTestFactory.GEBURTSNAME);
		}

		@Test
		void shouldSetGeburtsdatum() {
			antragstellerData.put("tfAntragstellerGeburtsdatum", Map.of(VALUE_KEY, AntragstellerTestFactory.GEBURTSDATUM, LABEL_KEY, "Geburtsdatum"));

			var result = buildAntragsteller();

			assertThat(result.getGeburtsdatum()).isEqualTo(AntragstellerTestFactory.GEBURTSDATUM);
		}

		@Test
		void shouldSetGeburtsort() {
			antragstellerData.put("tfAntragstellerGeburtsort", Map.of(VALUE_KEY, AntragstellerTestFactory.GEBURTSORT, LABEL_KEY, "Geburtsort"));

			var result = buildAntragsteller();

			assertThat(result.getGeburtsort()).isEqualTo(AntragstellerTestFactory.GEBURTSORT);
		}

		@Test
		void shouldSetEmail() {
			antragstellerData.put("tfAntragstellerEmail", Map.of(VALUE_KEY, AntragstellerTestFactory.EMAIL, LABEL_KEY, "E-Mail"));

			var result = buildAntragsteller();

			assertThat(result.getEmail()).isEqualTo(AntragstellerTestFactory.EMAIL);
		}

		@Test
		void shouldSetTelefon() {
			antragstellerData.put("tfAntragstellerTelefon", Map.of(VALUE_KEY, AntragstellerTestFactory.TELEFON, LABEL_KEY, "Telefon"));

			var result = buildAntragsteller();

			assertThat(result.getTelefon()).isEqualTo(AntragstellerTestFactory.TELEFON);
		}

		@Test
		void shouldSetAdresse() {
			antragstellerData.put("tfAntragstellerAdresse",
					Map.of(LABEL_KEY, "Adresse", VALUE_KEY, AntragstellerTestFactory.STRASSE + " " + AntragstellerTestFactory.HAUSNUMMER));

			var result = buildAntragsteller();

			assertThat(result.getStrasse()).isEqualTo(AntragstellerTestFactory.STRASSE + " " + AntragstellerTestFactory.HAUSNUMMER);
		}

		@Test
		void shouldSetPlz() {
			antragstellerData.put("tfAntragstellerPLZ", Map.of(LABEL_KEY, "PLZ", VALUE_KEY, AntragstellerTestFactory.PLZ));

			var result = buildAntragsteller();

			assertThat(result.getPlz()).isEqualTo(AntragstellerTestFactory.PLZ);
		}

		@Test
		void shouldSetOrt() {
			antragstellerData.put("tfAntragstellerOrt", Map.of(LABEL_KEY, "Ort", VALUE_KEY, AntragstellerTestFactory.ORT));

			var result = buildAntragsteller();

			assertThat(result.getOrt()).isEqualTo(AntragstellerTestFactory.ORT);
		}

		Antragsteller buildAntragsteller() {
			return mapper.buildAntragsteller(antragstellerData);
		}
	}

	@Nested
	class TestBuildFirmaAntragsteller {

		@BeforeEach
		void fillAntragstellerData() {
			antragstellerData.put("tfOrgName", Map.of(VALUE_KEY, AntragstellerTestFactory.FIRMA_NAME, LABEL_KEY, "Firmenname"));
			antragstellerData.put("tfAntragstellerAnrede", Map.of(VALUE_KEY, AntragstellerTestFactory.ANREDE, LABEL_KEY, "Anrede"));
			antragstellerData.put("tfAntragstellerVorname", Map.of(LABEL_KEY, "Vorname", VALUE_KEY, AntragstellerTestFactory.VORNAME));
			antragstellerData.put("tfAntragstellerName", Map.of(VALUE_KEY, AntragstellerTestFactory.NACHNAME, LABEL_KEY, "Nachname"));
			antragstellerData.put("tfAntragstellerGeburtsname", Map.of(VALUE_KEY, AntragstellerTestFactory.GEBURTSNAME, LABEL_KEY, "Geburtsname"));
			antragstellerData.put("tfAntragstellerGeburtsdatum", Map.of(VALUE_KEY, AntragstellerTestFactory.GEBURTSDATUM, LABEL_KEY, "Geburtsdatum"));
			antragstellerData.put("tfAntragstellerGeburtsort", Map.of(VALUE_KEY, AntragstellerTestFactory.GEBURTSORT, LABEL_KEY, "Geburtsort"));
			antragstellerData.put("tfAntragstellerEmail", Map.of(VALUE_KEY, AntragstellerTestFactory.EMAIL, LABEL_KEY, "E-Mail"));
			antragstellerData.put("tfAntragstellerTelefon", Map.of(VALUE_KEY, AntragstellerTestFactory.TELEFON, LABEL_KEY, "Telefon"));
			antragstellerData.put("tfAntragstellerAdresse",
					Map.of(LABEL_KEY, "Adresse", VALUE_KEY, AntragstellerTestFactory.STRASSE + " " + AntragstellerTestFactory.HAUSNUMMER));
			antragstellerData.put("tfAntragstellerPLZ", Map.of(LABEL_KEY, "PLZ", VALUE_KEY, AntragstellerTestFactory.PLZ));
			antragstellerData.put("tfAntragstellerOrt", Map.of(LABEL_KEY, "Ort", VALUE_KEY, AntragstellerTestFactory.ORT));
		}

		@Test
		void shouldMapFirmaData() {
			var expectedAntragsteller = AntragstellerTestFactory.createFirmBuilder()
					.strasse("%s %s".formatted(AntragstellerTestFactory.STRASSE, AntragstellerTestFactory.HAUSNUMMER))
					.hausnummer(null).postfachId(null).data(null).build();

			var result = mapper.buildAntragsteller(antragstellerData);

			assertThat(result).usingRecursiveComparison().isEqualTo(expectedAntragsteller);
		}

	}
}